"""Functions for finding group cohorts in chunked arrays."""

from __future__ import annotations

import itertools
import logging
import math
import operator
from concurrent.futures import ThreadPoolExecutor
from itertools import product
from typing import TYPE_CHECKING, Literal

import numpy as np
import toolz as tlz
from scipy.sparse import csc_array, csr_array

if TYPE_CHECKING:
    import pandas as pd

    from .core import T_Method

logger = logging.getLogger("flox")


def slices_from_chunks(chunks):
    """slightly modified from dask.array.core.slices_from_chunks to be lazy"""
    cumdims = [tlz.accumulate(operator.add, bds, 0) for bds in chunks]
    slices = (
        (slice(s, s + dim) for s, dim in zip(starts, shapes)) for starts, shapes in zip(cumdims, chunks)
    )
    return product(*slices)


def _compute_label_chunk_bitmask(labels, chunks, nlabels):
    def make_bitmask(rows, cols):
        data = np.broadcast_to(np.array(1, dtype=np.uint8), rows.shape)
        return csc_array((data, (rows, cols)), dtype=bool, shape=(nchunks, nlabels))

    assert isinstance(labels, np.ndarray)
    shape = tuple(sum(c) for c in chunks)
    nchunks = math.prod(len(c) for c in chunks)
    approx_chunk_size = math.prod(c[0] for c in chunks)

    # Shortcut for 1D with size-1 chunks
    if shape == (nchunks,):
        rows_array = np.arange(nchunks)
        cols_array = labels
        mask = labels >= 0
        return make_bitmask(rows_array[mask], cols_array[mask])

    labels = np.broadcast_to(labels, shape[-labels.ndim :])
    cols = []
    ilabels = np.arange(nlabels)

    def chunk_unique(labels, slicer, nlabels, label_is_present=None):
        if label_is_present is None:
            label_is_present = np.empty((nlabels + 1,), dtype=bool)
        label_is_present[:] = False
        subset = labels[slicer]
        # This is a quite fast way to find unique integers, when we know how many there are
        # inspired by a similar idea in numpy_groupies for first, last
        # instead of explicitly finding uniques, repeatedly write True to the same location
        label_is_present[subset.reshape(-1)] = True
        # skip the -1 sentinel by slicing
        # Faster than np.argwhere by a lot
        uniques = ilabels[label_is_present[:-1]]
        return uniques

    # TODO: refine this heuristic.
    # The general idea is that with the threadpool, we repeatedly allocate memory
    # for `label_is_present`. We trade that off against the parallelism across number of chunks.
    # For large enough number of chunks (relative to number of labels), it makes sense to
    # suffer the extra allocation in exchange for parallelism.
    THRESHOLD = 2
    if nlabels < THRESHOLD * approx_chunk_size:
        logger.debug(
            "Using threadpool since num_labels %s < %d * chunksize %s",
            nlabels,
            THRESHOLD,
            approx_chunk_size,
        )
        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(chunk_unique, labels, slicer, nlabels)
                for slicer in slices_from_chunks(chunks)
            ]
            cols = tuple(f.result() for f in futures)

    else:
        logger.debug(
            "Using serial loop since num_labels %s > %d * chunksize %s",
            nlabels,
            THRESHOLD,
            approx_chunk_size,
        )
        cols = []
        # Add one to handle the -1 sentinel value
        label_is_present = np.empty((nlabels + 1,), dtype=bool)
        for region in slices_from_chunks(chunks):
            uniques = chunk_unique(labels, region, nlabels, label_is_present)
            cols.append(uniques)
    rows_array = np.repeat(np.arange(nchunks), tuple(len(col) for col in cols))
    cols_array = np.concatenate(cols)

    return make_bitmask(rows_array, cols_array)


# @memoize
def find_group_cohorts(
    labels, chunks, expected_groups: pd.RangeIndex | None = None, merge: bool = False
) -> tuple[T_Method, dict]:
    """
    Finds groups labels that occur together aka "cohorts"

    If available, results are cached in a 1MB cache managed by `cachey`.
    This allows us to be quick when repeatedly calling groupby_reduce
    for arrays with the same chunking (e.g. an xarray Dataset).

    Parameters
    ----------
    labels : np.ndarray
        mD Array of integer group codes, factorized so that -1
        represents NaNs.
    chunks : tuple
        chunks of the array being reduced
    expected_groups: pd.RangeIndex (optional)
        Used to extract the largest label expected
    merge: bool (optional)
        Whether to merge cohorts or not. Set to True if a user
        specifies "cohorts" but other methods are preferable.

    Returns
    -------
    preferred_method: {"blockwise", cohorts", "map-reduce"}
    cohorts: dict_values
        Iterable of cohorts
    """
    # To do this, we must have values in memory so casting to numpy should be safe
    labels = np.asarray(labels)

    shape = tuple(sum(c) for c in chunks)
    nchunks = math.prod(len(c) for c in chunks)

    # assumes that `labels` are factorized
    if expected_groups is None:
        nlabels = labels.max() + 1
    else:
        nlabels = expected_groups[-1] + 1

    # 1. Single chunk, blockwise always
    if nchunks == 1:
        return "blockwise", {(0,): list(range(nlabels))}

    labels = np.broadcast_to(labels, shape[-labels.ndim :])
    bitmask = _compute_label_chunk_bitmask(labels, chunks, nlabels)

    CHUNK_AXIS, LABEL_AXIS = 0, 1
    chunks_per_label = bitmask.sum(axis=CHUNK_AXIS)

    # can happen when `expected_groups` is passed but not all labels are present
    # (binning, resampling)
    present_labels = np.arange(bitmask.shape[LABEL_AXIS])
    present_labels_mask = chunks_per_label != 0
    if not present_labels_mask.all():
        present_labels = present_labels[present_labels_mask]
        bitmask = bitmask[..., present_labels_mask]
        chunks_per_label = chunks_per_label[present_labels_mask]

    label_chunks = {
        present_labels[idx].item(): bitmask.indices[slice(bitmask.indptr[idx], bitmask.indptr[idx + 1])]
        for idx in range(bitmask.shape[LABEL_AXIS])
    }

    # Invert the label_chunks mapping so we know which labels occur together.
    def invert(x) -> tuple[np.ndarray, ...]:
        arr = label_chunks[x]
        return tuple(arr.tolist())

    chunks_cohorts = tlz.groupby(invert, label_chunks.keys())

    # 2. Every group is contained to one block, use blockwise here.
    if bitmask.shape[CHUNK_AXIS] == 1 or (chunks_per_label == 1).all():
        logger.debug("find_group_cohorts: blockwise is preferred.")
        return "blockwise", chunks_cohorts

    # 3. Perfectly chunked so there is only a single cohort
    if len(chunks_cohorts) == 1:
        logger.debug("Only found a single cohort. 'map-reduce' is preferred.")
        return "map-reduce", chunks_cohorts if merge else {}

    # 4. Our dataset has chunksize one along the axis,
    single_chunks = all(all(a == 1 for a in ac) for ac in chunks)
    # 5. Every chunk only has a single group, but that group might extend across multiple chunks
    one_group_per_chunk = (bitmask.sum(axis=LABEL_AXIS) == 1).all()
    # 6. Existing cohorts don't overlap, great for time grouping with perfect chunking
    no_overlapping_cohorts = (np.bincount(np.concatenate(tuple(chunks_cohorts.keys()))) == 1).all()
    if one_group_per_chunk or single_chunks or no_overlapping_cohorts:
        logger.debug("find_group_cohorts: cohorts is preferred, chunking is perfect.")
        return "cohorts", chunks_cohorts

    # We'll use containment to measure degree of overlap between labels.
    # Containment C = |Q & S| / |Q|
    #  - |X| is the cardinality of set X
    #  - Q is the query set being tested
    #  - S is the existing set
    # The bitmask matrix S allows us to calculate this pretty efficiently using a dot product.
    #       S.T @ S / chunks_per_label
    #
    # We treat the sparsity(C) = (nnz/size) as a summary measure of the net overlap.
    # 1. For high enough sparsity, there is a lot of overlap and we should use "map-reduce".
    # 2. When labels are uniformly distributed amongst all chunks
    #    (and number of labels < chunk size), sparsity is 1.
    # 3. Time grouping cohorts (e.g. dayofyear) appear as lines in this matrix.
    # 4. When there are no overlaps at all between labels, containment is a block diagonal matrix
    #    (approximately).
    #
    # However computing S.T @ S can still be the slowest step, especially if S
    # is not particularly sparse. Empirically the sparsity( S.T @ S ) > min(1, 2 x sparsity(S)).
    # So we use sparsity(S) as a shortcut.
    MAX_SPARSITY_FOR_COHORTS = 0.4  # arbitrary
    sparsity = bitmask.nnz / math.prod(bitmask.shape)
    preferred_method: Literal["map-reduce"] | Literal["cohorts"]
    logger.debug(
        "sparsity of bitmask is {}, threshold is {}".format(  # noqa
            sparsity, MAX_SPARSITY_FOR_COHORTS
        )
    )
    # 7. Groups seem fairly randomly distributed, use "map-reduce".
    if sparsity > MAX_SPARSITY_FOR_COHORTS:
        if not merge:
            logger.debug(
                "find_group_cohorts: bitmask sparsity={}, merge=False, choosing 'map-reduce'".format(  # noqa
                    sparsity
                )
            )
            return "map-reduce", {}
        preferred_method = "map-reduce"
    else:
        preferred_method = "cohorts"

    # Note: While A.T @ A is a symmetric matrix, the division by chunks_per_label
    #       makes it non-symmetric.
    asfloat = bitmask.astype(float)
    containment = csr_array(asfloat.T @ asfloat / chunks_per_label)

    logger.debug(
        "sparsity of containment matrix is {}".format(  # noqa
            containment.nnz / math.prod(containment.shape)
        )
    )

    # Next we for-loop over groups and merge those that are quite similar.
    # Use a threshold on containment to always force some merging.
    # Note that we do not use the filtered containment matrix for estimating "sparsity"
    # because it is a bit hard to reason about.
    MIN_CONTAINMENT = 0.75  # arbitrary
    mask = containment.data < MIN_CONTAINMENT

    # Now we also know "exact cohorts" -- cohorts whose constituent groups
    # occur in exactly the same chunks. We only need examine one member of each group.
    # Skip the others by first looping over the exact cohorts, and zero out those rows.
    repeated = np.concatenate([v[1:] for v in chunks_cohorts.values()]).astype(int)
    repeated_idx = np.searchsorted(present_labels, repeated)
    for i in repeated_idx:
        mask[containment.indptr[i] : containment.indptr[i + 1]] = True
    containment.data[mask] = 0
    containment.eliminate_zeros()

    # Figure out all the labels we need to loop over later
    n_overlapping_labels = containment.astype(bool).sum(axis=1)
    order = np.argsort(n_overlapping_labels, kind="stable")[::-1]
    # Order is such that we iterate over labels, beginning with those with most overlaps
    # Also filter out any "exact" cohorts
    order = order[n_overlapping_labels[order] > 0]

    logger.debug("find_group_cohorts: merging cohorts")
    merged_cohorts = {}
    merged_keys = set()
    for rowidx in order:
        if present_labels[rowidx] in merged_keys:
            continue
        cohidx = containment.indices[slice(containment.indptr[rowidx], containment.indptr[rowidx + 1])]
        cohort_ = present_labels[cohidx]
        cohort = [elem.item() for elem in cohort_ if elem not in merged_keys]
        if not cohort:
            continue
        merged_keys.update(cohort)
        allchunks = (label_chunks[member].tolist() for member in cohort)
        chunk = tuple(set(itertools.chain(*allchunks)))
        merged_cohorts[chunk] = cohort

    actual_ngroups = np.concatenate(tuple(merged_cohorts.values())).size
    expected_ngroups = present_labels.size
    assert len(merged_keys) == actual_ngroups
    assert expected_ngroups == actual_ngroups, (expected_ngroups, actual_ngroups)

    # sort by first label in cohort
    # This will help when sort=True (default)
    # and we have to resort the dask array
    as_sorted = dict(sorted(merged_cohorts.items(), key=lambda kv: kv[1][0]))
    return preferred_method, as_sorted



from collections.abc import Callable
from typing import Self

import numpy as np

MULTIARRAY_HANDLED_FUNCTIONS: dict[Callable, Callable] = {}


class MultiArray:
    arrays: tuple[np.ndarray, ...]

    def __init__(self, arrays):
        self.arrays = arrays
        assert all(arrays[0].shape == a.shape for a in arrays), "Expect all arrays to have the same shape"

    def astype(self, dt, **kwargs) -> Self:
        return type(self)(tuple(array.astype(dt, **kwargs) for array in self.arrays))

    def reshape(self, shape, **kwargs) -> Self:
        return type(self)(tuple(array.reshape(shape, **kwargs) for array in self.arrays))

    def squeeze(self, axis=None) -> Self:
        return type(self)(tuple(array.squeeze(axis) for array in self.arrays))

    def __setitem__(self, key, value) -> None:
        assert len(value) == len(self.arrays)
        for array, val in zip(self.arrays, value):
            array[key] = val

    def __array_function__(self, func, types, args, kwargs):
        if func not in MULTIARRAY_HANDLED_FUNCTIONS:
            return NotImplemented
        # Note: this allows subclasses that don't override
        # __array_function__ to handle MyArray objects
        # if not all(issubclass(t, MyArray) for t in types): # I can't see this being relevant at all for this code, but maybe it's safer to leave it in?
        # return NotImplemented
        return MULTIARRAY_HANDLED_FUNCTIONS[func](*args, **kwargs)

    # Shape is needed, seems likely that the other two might be
    # Making some strong assumptions here that all the arrays are the same shape, and I don't really like this
    @property
    def dtype(self) -> np.dtype:
        return self.arrays[0].dtype

    @property
    def shape(self) -> tuple[int, ...]:
        return self.arrays[0].shape

    @property
    def ndim(self) -> int:
        return self.arrays[0].ndim

    def __getitem__(self, key) -> Self:
        return type(self)([array[key] for array in self.arrays])


def implements(numpy_function):
    """Register an __array_function__ implementation for MyArray objects."""

    def decorator(func):
        MULTIARRAY_HANDLED_FUNCTIONS[numpy_function] = func
        return func

    return decorator


@implements(np.expand_dims)
def expand_dims(multiarray, axis) -> MultiArray:
    return MultiArray(tuple(np.expand_dims(a, axis) for a in multiarray.arrays))


@implements(np.concatenate)
def concatenate(multiarrays, axis) -> MultiArray:
    n_arrays = len(multiarrays[0].arrays)
    for ma in multiarrays[1:]:
        assert len(ma.arrays) == n_arrays
    return MultiArray(
        tuple(np.concatenate(tuple(ma.arrays[i] for ma in multiarrays), axis) for i in range(n_arrays))
    )


@implements(np.transpose)
def transpose(multiarray, axes) -> MultiArray:
    return MultiArray(tuple(np.transpose(a, axes) for a in multiarray.arrays))


@implements(np.squeeze)
def squeeze(multiarray, axis) -> MultiArray:
    return MultiArray(tuple(np.squeeze(a, axis) for a in multiarray.arrays))


@implements(np.full)
def full(shape, fill_values, *args, **kwargs) -> MultiArray:
    """All arguments except fill_value are shared by each array in the MultiArray.
    Iterate over fill_values to create arrays
    """
    return MultiArray(tuple(np.full(shape, fv, *args, **kwargs) for fv in fill_values))



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, TypeAlias, TypeVar

from .types import DaskArray, Graph
from .xrutils import is_duck_dask_array, module_available

if TYPE_CHECKING:
    from .aggregations import Aggregation

    T_Agg: TypeAlias = str | Aggregation

T = TypeVar("T")

try:
    import dask.array as da

    dask_array_type = da.Array
except ImportError:
    dask_array_type = ()  # type: ignore[assignment, misc]

try:
    import sparse

    sparse_array_type = sparse.COO
except ImportError:
    sparse_array_type = ()

HAS_SPARSE = module_available("sparse")


@dataclass
class ArrayLayer:
    name: str
    layer: Graph
    chunks: tuple[tuple[int, ...], ...]

    def to_array(self, dep: DaskArray) -> DaskArray:
        from dask.array import Array
        from dask.highlevelgraph import HighLevelGraph

        graph = HighLevelGraph.from_collections(self.name, self.layer, dependencies=[dep])
        return Array(graph, self.name, self.chunks, meta=dep._meta)


def _postprocess_numbagg(result, *, func, fill_value, size, seen_groups):
    """Account for numbagg not providing a fill_value kwarg."""
    import numpy as np

    from .aggregate_numbagg import DEFAULT_FILL_VALUE

    if not isinstance(func, str) or func not in DEFAULT_FILL_VALUE:
        return result
    # The condition needs to be
    # len(found_groups) < size; if so we mask with fill_value (?)
    default_fv = DEFAULT_FILL_VALUE[func]
    needs_masking = fill_value is not None and not np.array_equal(fill_value, default_fv, equal_nan=True)
    groups = np.arange(size)
    if needs_masking:
        mask = np.isin(groups, seen_groups, assume_unique=True, invert=True)
        if mask.any():
            if isinstance(result, sparse_array_type):
                result.fill_value = fill_value
            else:
                result[..., groups[mask]] = fill_value
    return result


def identity(x: T) -> T:
    return x


def _issorted(arr, ascending=True) -> bool:
    if ascending:
        return bool((arr[:-1] <= arr[1:]).all())
    else:
        return bool((arr[:-1] >= arr[1:]).all())


def _should_auto_rechunk_blockwise(method, array, any_by_dask: bool, by) -> bool:
    """Check if we should attempt automatic rechunking for blockwise operations."""
    return method is None and is_duck_dask_array(array) and not any_by_dask and by.ndim == 1 and _issorted(by)


def _is_nanlen(reduction) -> bool:
    return isinstance(reduction, str) and reduction == "nanlen"


def _is_arg_reduction(func: T_Agg) -> bool:
    from .aggregations import Aggregation

    if isinstance(func, str) and func in ["argmin", "argmax", "nanargmax", "nanargmin"]:
        return True
    if isinstance(func, Aggregation) and func.reduction_type == "argreduce":
        return True
    return False


def _is_minmax_reduction(func: T_Agg) -> bool:
    return not _is_arg_reduction(func) and (isinstance(func, str) and ("max" in func or "min" in func))


def _is_first_last_reduction(func: T_Agg) -> bool:
    from .aggregations import Aggregation

    if isinstance(func, Aggregation):
        func = func.name
    return func in ["nanfirst", "nanlast", "first", "last"]


def _is_bool_supported_reduction(func: T_Agg) -> bool:
    from .aggregations import Aggregation

    if isinstance(func, Aggregation):
        func = func.name
    return (
        func in ["all", "any"]
        # TODO: enable in npg
        # or _is_first_last_reduction(func)
        # or _is_minmax_reduction(func)
    )


def _is_sparse_supported_reduction(func: T_Agg) -> bool:
    from .aggregations import SCANS, Aggregation

    if isinstance(func, Aggregation):
        func = func.name
    if func in SCANS:
        return False
    return not _is_arg_reduction(func) and any(f in func for f in ["len", "sum", "max", "min", "mean"])


def _is_reindex_sparse_supported_reduction(func: T_Agg) -> bool:
    from .aggregations import Aggregation

    if isinstance(func, Aggregation):
        func = func.name
    return HAS_SPARSE and all(f not in func for f in ["first", "last", "prod", "var", "std"])



#!/usr/bin/env python
# flake8: noqa
"""Top-level module for flox ."""

from . import cache, cohorts, factorize
from .aggregations import Aggregation, Scan, is_supported_aggregation
from .core import groupby_reduce
from .options import set_options
from .rechunk import rechunk_for_blockwise, rechunk_for_cohorts
from .reindex import ReindexArrayType, ReindexStrategy
from .scan import groupby_scan


def _get_version():
    __version__ = "999"
    try:
        from ._version import __version__
    except ImportError:
        pass
    return __version__


__version__ = _get_version()

__all__ = [
    "Aggregation",
    "Scan",
    "groupby_reduce",
    "groupby_scan",
    "rechunk_for_blockwise",
    "rechunk_for_cohorts",
    "set_options",
    "ReindexStrategy",
    "ReindexArrayType",
    "is_supported_aggregation",
]



"""Scan operations for groupby reductions.

This module provides scan operations (cumulative reductions) for grouped data.
"""

from __future__ import annotations

import copy
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from .aggregations import (
    SCANS,
    AlignedArrays,
    Scan,
    ScanState,
    _atleast_1d,
    generic_aggregate,
)
from .cohorts import find_group_cohorts
from .factorize import _factorize_multiple
from .lib import _should_auto_rechunk_blockwise
from .rechunk import rechunk_for_blockwise
from .xrutils import is_duck_array, is_duck_dask_array, module_available

if module_available("numpy", minversion="2.0.0"):
    from numpy.lib.array_utils import normalize_axis_tuple
else:
    from numpy.core.numeric import normalize_axis_tuple  # type: ignore[no-redef]

if TYPE_CHECKING:
    from typing import Literal

    from .core import (
        T_By,
        T_EngineOpt,
        T_ExpectedGroupsOpt,
        T_MethodOpt,
        T_Scan,
    )
    from .types import DaskArray

    T_ScanMethod = Literal["blockwise", "blelloch"]


def _choose_scan_method(
    method: T_MethodOpt, preferred_method: T_ScanMethod, nax: int, by_ndim: int
) -> T_ScanMethod:
    """Choose the scan method based on user input and preferred method.

    Parameters
    ----------
    method : T_MethodOpt
        User-specified method, or None for automatic selection.
    preferred_method : T_ScanMethod
        The preferred method based on data layout analysis.
    nax : int
        Number of axes being reduced.
    by_ndim : int
        Number of dimensions in the `by` array.

    Returns
    -------
    T_ScanMethod
        The chosen scan method: "blockwise" or "blelloch".
    """
    if method is None:
        # Scans must reduce along all dimensions of by for blockwise
        if nax != by_ndim:
            return "blelloch"
        return preferred_method
    elif method == "blockwise":
        return "blockwise"
    else:
        # For any other method (including map-reduce, cohorts), use blelloch
        return "blelloch"


def _validate_expected_groups(nby, expected_groups):
    """Validate expected_groups for scan operations."""
    if expected_groups is None:
        return (None,) * nby
    return expected_groups


def _convert_expected_groups_to_index(expected_groups):
    """Convert expected_groups to index for scan operations."""
    result = []
    for expect in expected_groups:
        if expect is None:
            result.append(None)
        elif isinstance(expect, pd.Index):
            result.append(expect)
        else:
            result.append(pd.Index(expect))
    return tuple(result)


def groupby_scan(
    array: np.ndarray | DaskArray,
    *by: T_By,
    func: T_Scan,
    expected_groups: T_ExpectedGroupsOpt = None,
    axis: int | tuple[int] = -1,
    dtype: np.typing.DTypeLike = None,
    method: T_MethodOpt = None,
    engine: T_EngineOpt = None,
) -> np.ndarray | DaskArray:
    """
    GroupBy reductions using parallel scans for dask.array

    Parameters
    ----------
    array : ndarray or DaskArray
        Array to be reduced, possibly nD
    *by : ndarray or DaskArray
        Array of labels to group over. Must be aligned with ``array`` so that
        ``array.shape[-by.ndim :] == by.shape`` or any disagreements in that
        equality check are for dimensions of size 1 in `by`.
    func : {"nancumsum", "ffill", "bfill"} or Scan
        Single function name or a Scan instance
    expected_groups : (optional) Sequence
        Expected unique labels.
    axis : None or int or Sequence[int], optional
        If None, reduce across all dimensions of by
        Else, reduce across corresponding axes of array
        Negative integers are normalized using array.ndim.
    fill_value : Any
        Value to assign when a label in ``expected_groups`` is not present.
    dtype : data-type , optional
        DType for the output. Can be anything that is accepted by ``np.dtype``.
    method : {"blockwise", "blelloch"}, optional
        Strategy for scan of dask arrays only:
          * ``"blockwise"``:
            Only scan using blockwise and avoid aggregating blocks
            together. Useful for resampling-style groupby problems where group
            members are always together. If  `by` is 1D,  `array` is automatically
            rechunked so that chunk boundaries line up with group boundaries
            i.e. each block contains all members of any group present
            in that block. For nD `by`, you must make sure that all members of a group
            are present in a single block.
          * ``"blelloch"``:
            Use Blelloch's parallel prefix scan algorithm, which allows
            scanning across chunk boundaries. This is the default when groups
            span multiple chunks.
    engine : {"flox", "numpy", "numba", "numbagg"}, optional
        Algorithm to compute the groupby reduction on non-dask arrays and on each dask chunk:
          * ``"numpy"``:
            Use the vectorized implementations in ``numpy_groupies.aggregate_numpy``.
            This is the default choice because it works for most array types.
          * ``"flox"``:
            Use an internal implementation where the data is sorted so that
            all members of a group occur sequentially, and then numpy.ufunc.reduceat
            is to used for the reduction. This will fall back to ``numpy_groupies.aggregate_numpy``
            for a reduction that is not yet implemented.
          * ``"numba"``:
            Use the implementations in ``numpy_groupies.aggregate_numba``.
          * ``"numbagg"``:
            Use the reductions supported by ``numbagg.grouped``. This will fall back to ``numpy_groupies.aggregate_numpy``
            for a reduction that is not yet implemented.

    Returns
    -------
    result
        Aggregated result

    See Also
    --------
    xarray.xarray_reduce
    """
    from . import xrdtypes

    axis = _atleast_1d(axis)
    if len(axis) > 1:
        raise NotImplementedError("Scans are only supported along a single dimension.")

    bys: tuple = tuple(np.asarray(b) if not is_duck_array(b) else b for b in by)
    nby = len(by)
    by_is_dask = tuple(is_duck_dask_array(b) for b in bys)
    any_by_dask = any(by_is_dask)

    axis_ = normalize_axis_tuple(axis, array.ndim)

    if engine is not None:
        raise NotImplementedError("Setting `engine` is not supported for scans yet.")
    if engine is None:
        engine = "flox"
    assert engine == "flox"

    if not is_duck_array(array):
        array = np.asarray(array)

    agg = SCANS[func] if isinstance(func, str) else func
    assert isinstance(agg, Scan)
    agg = copy.deepcopy(agg)

    if (agg == SCANS["ffill"] or agg == SCANS["bfill"]) and array.dtype.kind != "f":
        # nothing to do, no NaNs!
        return array

    if expected_groups is not None:
        raise NotImplementedError("Setting `expected_groups` with scans is not supported yet.")
    expected_groups = _validate_expected_groups(nby, expected_groups)
    expected_groups = _convert_expected_groups_to_index(expected_groups)

    # Don't factorize early only when
    # grouping by dask arrays, and not having expected_groups
    factorize_early = not (
        # can't do it if we are grouping by dask array but don't have expected_groups
        any(is_dask and ex_ is None for is_dask, ex_ in zip(by_is_dask, expected_groups))
    )
    if factorize_early:
        bys, final_groups, grp_shape = _factorize_multiple(
            bys,
            expected_groups,
            any_by_dask=any_by_dask,
            sort=False,
        )
    else:
        raise NotImplementedError

    assert len(bys) == 1
    by_: np.ndarray
    (by_,) = bys
    has_dask = is_duck_dask_array(array) or is_duck_dask_array(by_)
    nax = len(axis_)

    # Method selection for dask arrays
    scan_method: T_ScanMethod = "blelloch"
    if has_dask:
        (single_axis,) = axis_  # type: ignore[misc]

        # Try rechunking for sorted 1D by when method is not specified
        if _should_auto_rechunk_blockwise(method, array, any_by_dask, by_):
            rechunk_method, array = rechunk_for_blockwise(array, single_axis, by_, force=False)
            if rechunk_method == "blockwise":
                method = "blockwise"

        # Determine preferred method based on data layout
        if not any_by_dask and method is None:
            cohorts_method, _ = find_group_cohorts(
                by_,
                [array.chunks[ax] for ax in range(-by_.ndim, 0)],  # type: ignore[union-attr]
                expected_groups=None,
                merge=False,
            )
            # Map groupby_reduce methods to scan methods
            preferred_method: T_ScanMethod = "blockwise" if cohorts_method == "blockwise" else "blelloch"
        else:
            preferred_method = "blelloch"

        # Choose the final method
        scan_method = _choose_scan_method(method, preferred_method, nax, by_.ndim)

        # Rechunk if blockwise was explicitly requested but data isn't aligned
        if preferred_method != "blockwise" and scan_method == "blockwise" and by_.ndim == 1:
            _, array = rechunk_for_blockwise(array, axis=-1, labels=by_)

    if array.dtype.kind in "Mm":
        cast_to = array.dtype
        array = array.view(np.int64)
    elif array.dtype.kind == "b":
        array = array.view(np.int8)
        cast_to = None
        if agg.preserves_dtype:
            cast_to = bool
    else:
        cast_to = None

    # TODO: move to aggregate_npg.py
    if agg.name in ["cumsum", "nancumsum"] and array.dtype.kind in ["i", "u"]:
        # https://numpy.org/doc/stable/reference/generated/numpy.cumsum.html
        # it defaults to the dtype of a, unless a
        # has an integer dtype with a precision less than that of the default platform integer.
        if array.dtype.kind == "i":
            agg.dtype = np.result_type(array.dtype, np.int_)
        elif array.dtype.kind == "u":
            agg.dtype = np.result_type(array.dtype, np.uint)
    else:
        agg.dtype = array.dtype if dtype is None else dtype
    agg.identity = xrdtypes._get_fill_value(agg.dtype, agg.identity)

    (single_axis,) = axis_  # type: ignore[misc]
    # avoid some roundoff error when we can.
    if by_.shape[-1] == 1 or by_.shape == grp_shape:
        array = array.astype(agg.dtype)
        if cast_to is not None:
            array = array.astype(cast_to)
        return array

    # Made a design choice here to have `preprocess` handle both array and group_idx
    # Example: for reversing, we need to reverse the whole array, not just reverse
    #          each block independently
    inp = AlignedArrays(array=array, group_idx=by_)
    if agg.preprocess:
        inp = agg.preprocess(inp)

    if not has_dask:
        final_state = chunk_scan(inp, axis=single_axis, agg=agg, dtype=agg.dtype)
        result = _finalize_scan(final_state, dtype=agg.dtype)
    else:
        from .dask import dask_groupby_scan

        result = dask_groupby_scan(inp.array, inp.group_idx, axes=axis_, agg=agg, method=scan_method)

    # Made a design choice here to have `postprocess` handle both array and group_idx
    out = AlignedArrays(array=result, group_idx=by_)
    if agg.finalize:
        out = agg.finalize(out)

    if cast_to is not None:
        return out.array.astype(cast_to)
    return out.array


def chunk_scan(inp: AlignedArrays, *, axis: int, agg: Scan, dtype=None, keepdims=None) -> ScanState:
    assert axis == inp.array.ndim - 1

    # I don't think we need to re-factorize here unless we are grouping by a dask array
    accumulated = generic_aggregate(
        inp.group_idx,
        inp.array,
        axis=axis,
        engine="flox",
        func=agg.scan,
        dtype=dtype,
        fill_value=agg.identity,
    )
    result = AlignedArrays(array=accumulated, group_idx=inp.group_idx)
    return ScanState(result=result, state=None)


def grouped_reduce(inp: AlignedArrays, *, agg: Scan, axis: int, keepdims=None) -> ScanState:
    from .core import chunk_reduce

    assert axis == inp.array.ndim - 1
    reduced = chunk_reduce(
        inp.array,
        inp.group_idx,
        func=(agg.reduction,),
        axis=axis,
        engine="flox",
        dtype=inp.array.dtype,
        fill_value=agg.identity,
        expected_groups=None,
    )
    return ScanState(
        state=AlignedArrays(array=reduced["intermediates"][0], group_idx=reduced["groups"]),
        result=None,
    )


def _zip(group_idx: np.ndarray, array: np.ndarray) -> AlignedArrays:
    return AlignedArrays(group_idx=group_idx, array=array)


def _finalize_scan(block: ScanState, dtype) -> np.ndarray:
    assert block.result is not None
    return block.result.array.astype(dtype, copy=False)


__all__ = [
    "_finalize_scan",
    "_zip",
    "chunk_scan",
    "grouped_reduce",
    "groupby_scan",
]



"""Factorization functions for groupby operations.

This module provides functions for factorizing groupby labels.
"""

from __future__ import annotations

import itertools
import math
from concurrent.futures import ThreadPoolExecutor
from functools import partial, reduce
from typing import TYPE_CHECKING, Literal, cast, overload

import numpy as np
import pandas as pd

from .types import FactorizeKwargs, FactorProps
from .xrutils import is_duck_dask_array, isnull

if TYPE_CHECKING:
    from .core import T_Axes, T_By, T_Bys, T_ExpectIndexOptTuple


def offset_labels(labels: np.ndarray, ngroups: int) -> tuple[np.ndarray, int]:
    """
    Offset group labels by dimension. This is used when we
    reduce over a subset of the dimensions of by. It assumes that the reductions
    dimensions have been flattened in the last dimension
    Copied from xhistogram &
    https://stackoverflow.com/questions/46256279/bin-elements-per-row-vectorized-2d-bincount-for-numpy
    """
    assert labels.ndim > 1
    offset: np.ndarray = (
        labels + np.arange(math.prod(labels.shape[:-1])).reshape((*labels.shape[:-1], -1)) * ngroups
    )
    # -1 indicates NaNs. preserve these otherwise we aggregate in the wrong groups!
    offset[labels == -1] = -1
    size: int = math.prod(labels.shape[:-1]) * ngroups
    return offset, size


def _factorize_single(by, expect, *, sort: bool, reindex: bool) -> tuple[pd.Index, np.ndarray]:
    flat = by.reshape(-1)
    if isinstance(expect, pd.RangeIndex):
        # idx is a view of the original `by` array
        # copy here so we don't have a race condition with the
        # group_idx[nanmask] = nan_sentinel assignment later
        # this is important in shared-memory parallelism with dask
        # TODO: figure out how to avoid this
        idx = flat.copy()
        found_groups = cast(pd.Index, expect)
        # TODO: fix by using masked integers
        idx[idx > expect[-1]] = -1

    elif isinstance(expect, pd.IntervalIndex):
        if expect.closed == "both":
            raise NotImplementedError
        bins = np.concatenate([expect.left.to_numpy(), expect.right.to_numpy()[[-1]]])

        # digitize is 0 or idx.max() for values outside the bounds of all intervals
        # make it behave like pd.cut which uses -1:
        if len(bins) > 1:
            right = expect.closed_right
            # For datetime64, convert both flat and bins to int64 for comparison
            is_datetime = bins.dtype.kind == "M"
            flat_for_digitize = flat.view(np.int64) if is_datetime else flat
            bins_for_digitize = bins.view(np.int64) if is_datetime else bins
            idx = np.digitize(
                flat_for_digitize,
                bins=bins_for_digitize,
                right=right,
            )
            idx -= 1
            within_bins = (
                flat_for_digitize <= bins_for_digitize.max()
                if right
                else flat_for_digitize < bins_for_digitize.max()
            )
            idx[~within_bins] = -1
        else:
            idx = np.zeros_like(flat, dtype=np.intp) - 1
        found_groups = cast(pd.Index, expect)
    else:
        if expect is not None and reindex:
            sorter = np.argsort(expect)
            groups = expect[sorter] if sort else expect
            idx = np.searchsorted(expect, flat, sorter=sorter)
            mask = ~np.isin(flat, expect) | isnull(flat) | (idx == len(expect))
            if not sort:
                # idx is the index in to the sorted array.
                # if we didn't want sorting, unsort it back
                idx[(idx == len(expect),)] = -1
                idx = sorter[(idx,)]
            idx[mask] = -1
        else:
            idx, groups = pd.factorize(flat, sort=sort)
        found_groups = cast(pd.Index, groups)

    return (found_groups, idx.reshape(by.shape))


def _ravel_factorized(*factorized: np.ndarray, grp_shape: tuple[int, ...]) -> np.ndarray:
    group_idx = np.ravel_multi_index(factorized, grp_shape, mode="wrap")
    # NaNs; as well as values outside the bins are coded by -1
    # Restore these after the raveling
    nan_by_mask = reduce(np.logical_or, [(f == -1) for f in factorized])
    group_idx[nan_by_mask] = -1
    return group_idx


@overload
def factorize_(
    by: T_Bys,
    axes: T_Axes,
    *,
    fastpath: Literal[True],
    expected_groups: T_ExpectIndexOptTuple | None = None,
    reindex: bool = False,
    sort: bool = True,
) -> tuple[np.ndarray, tuple[pd.Index, ...], tuple[int, ...], int, int, None]: ...


@overload
def factorize_(
    by: T_Bys,
    axes: T_Axes,
    *,
    expected_groups: T_ExpectIndexOptTuple | None = None,
    reindex: bool = False,
    sort: bool = True,
    fastpath: Literal[False] = False,
) -> tuple[np.ndarray, tuple[pd.Index, ...], tuple[int, ...], int, int, FactorProps]: ...


@overload
def factorize_(
    by: T_Bys,
    axes: T_Axes,
    *,
    expected_groups: T_ExpectIndexOptTuple | None = None,
    reindex: bool = False,
    sort: bool = True,
    fastpath: bool = False,
) -> tuple[np.ndarray, tuple[pd.Index, ...], tuple[int, ...], int, int, FactorProps | None]: ...


def factorize_(
    by: T_Bys,
    axes: T_Axes,
    *,
    expected_groups: T_ExpectIndexOptTuple | None = None,
    reindex: bool = False,
    sort: bool = True,
    fastpath: bool = False,
) -> tuple[np.ndarray, tuple[pd.Index, ...], tuple[int, ...], int, int, FactorProps | None]:
    """
    Returns an array of integer codes for groups (and associated data)
    by wrapping pd.cut and pd.factorize (depending on isbin).
    This method handles reindex and sort so that we don't spend time reindexing / sorting
    a possibly large results array. Instead we set up the appropriate integer codes (group_idx)
    so that the results come out in the appropriate order.
    """
    if expected_groups is None:
        expected_groups = (None,) * len(by)

    if len(by) > 2:
        with ThreadPoolExecutor() as executor:
            futures = [
                executor.submit(partial(_factorize_single, sort=sort, reindex=reindex), groupvar, expect)
                for groupvar, expect in zip(by, expected_groups)
            ]
            results = tuple(f.result() for f in futures)
    else:
        results = tuple(
            _factorize_single(groupvar, expect, sort=sort, reindex=reindex)
            for groupvar, expect in zip(by, expected_groups)
        )
    found_groups = tuple(r[0] for r in results)
    factorized = [r[1] for r in results]

    grp_shape = tuple(len(grp) for grp in found_groups)
    ngroups = math.prod(grp_shape)
    if len(by) > 1:
        group_idx = _ravel_factorized(*factorized, grp_shape=grp_shape)
    else:
        (group_idx,) = factorized

    if fastpath:
        return group_idx, found_groups, grp_shape, ngroups, ngroups, None

    if len(axes) == 1 and by[0].ndim > 1:
        # Not reducing along all dimensions of by
        # this is OK because for 3D by and axis=(1,2),
        # we collapse to a 2D by and axis=-1
        offset_group = True
        group_idx, size = offset_labels(group_idx.reshape(by[0].shape), ngroups)
    else:
        size = ngroups
        offset_group = False

    # numpy_groupies cannot deal with group_idx = -1
    # so we'll add use ngroups as the sentinel
    # note we cannot simply remove the NaN locations;
    # that would mess up argmax, argmin
    nan_sentinel = size if offset_group else ngroups
    nanmask = group_idx == -1
    if nanmask.any():
        # bump it up so there's a place to assign values to the nan_sentinel index
        size += 1
    group_idx[nanmask] = nan_sentinel

    props = FactorProps(offset_group, nan_sentinel, nanmask)
    return group_idx, tuple(found_groups), grp_shape, ngroups, size, props


def _lazy_factorize_wrapper(*by: T_By, **kwargs) -> np.ndarray:
    group_idx, *_ = factorize_(by, **kwargs)
    return group_idx


def _factorize_multiple(
    by: T_Bys,
    expected_groups: T_ExpectIndexOptTuple,
    any_by_dask: bool,
    sort: bool = True,
) -> tuple[tuple[np.ndarray], tuple[pd.Index, ...], tuple[int, ...]]:
    kwargs: FactorizeKwargs = dict(
        axes=(),  # always (), we offset later if necessary.
        fastpath=True,
        # This is the only way it makes sense I think.
        # reindex controls what's actually allocated in chunk_reduce
        # At this point, we care about an accurate conversion to codes.
        reindex=True,
        sort=sort,
    )
    if any_by_dask:
        import dask.array

        from . import dask_array_ops  # noqa

        # unifying chunks will make sure all arrays in `by` are dask arrays
        # with compatible chunks, even if there was originally a numpy array
        inds = tuple(range(by[0].ndim))
        for by_, expect in zip(by, expected_groups):
            if expect is None and is_duck_dask_array(by_):
                raise ValueError("Please provide expected_groups when grouping by a dask array.")

        found_groups = tuple(
            pd.Index(pd.unique(by_.reshape(-1))) if expect is None else expect
            for by_, expect in zip(by, expected_groups)
        )
        grp_shape = tuple(map(len, found_groups))

        chunks, by_chunked = dask.array.unify_chunks(*itertools.chain(*zip(by, (inds,) * len(by))))
        group_idxs = [
            dask.array.map_blocks(
                _lazy_factorize_wrapper,
                by_,
                expected_groups=(expect_,),
                meta=np.array((), dtype=np.int64),
                **kwargs,
            )
            for by_, expect_ in zip(by_chunked, expected_groups)
        ]
        # This could be avoied but we'd use `np.where`
        # instead `_ravel_factorized` instead i.e. a copy.
        group_idx = dask.array.map_blocks(
            _ravel_factorized, *group_idxs, grp_shape=grp_shape, chunks=tuple(chunks.values()), dtype=np.int64
        )

    else:
        kwargs["by"] = by
        group_idx, found_groups, grp_shape, *_ = factorize_(**kwargs, expected_groups=expected_groups)

    return (group_idx,), found_groups, grp_shape


__all__ = [
    "FactorizeKwargs",
    "FactorProps",
    "_factorize_multiple",
    "_factorize_single",
    "_lazy_factorize_wrapper",
    "_ravel_factorized",
    "factorize_",
    "offset_labels",
]



"""Rechunking functions for groupby operations.

This module provides functions for rechunking arrays to optimize groupby operations.
"""

from __future__ import annotations

import logging
from collections.abc import Sequence
from typing import TYPE_CHECKING

import numpy as np
import numpy_groupies as npg
import pandas as pd

from .aggregations import _atleast_1d
from .cache import memoize
from .factorize import factorize_
from .options import OPTIONS

if TYPE_CHECKING:
    from .core import T_Axis, T_MethodOpt
    from .types import DaskArray

logger = logging.getLogger("flox")


@memoize
def _get_optimal_chunks_for_groups(chunks, labels):
    chunkidx = np.cumsum(chunks) - 1
    # what are the groups at chunk boundaries
    labels_at_chunk_bounds = pd.unique(labels[chunkidx])
    # what's the last index of all groups
    last_indexes = npg.aggregate_numpy.aggregate(labels, np.arange(len(labels)), func="last")
    # what's the last index of groups at the chunk boundaries.
    lastidx = last_indexes[labels_at_chunk_bounds]

    if len(chunkidx) == len(lastidx) and (chunkidx == lastidx).all():
        return chunks

    first_indexes = npg.aggregate_numpy.aggregate(labels, np.arange(len(labels)), func="first")
    firstidx = first_indexes[labels_at_chunk_bounds]

    newchunkidx = [0]
    for c, f, l in zip(chunkidx, firstidx, lastidx):  # noqa
        Δf = abs(c - f)
        Δl = abs(c - l)
        if c == 0 or newchunkidx[-1] > l:
            continue
        f = f.item()  # noqa
        l = l.item()  # noqa
        if Δf < Δl and f > newchunkidx[-1]:
            newchunkidx.append(f)
        else:
            newchunkidx.append(l + 1)
    if newchunkidx[-1] != chunkidx[-1] + 1:
        newchunkidx.append(chunkidx[-1] + 1)
    newchunks = np.diff(newchunkidx)

    assert sum(newchunks) == sum(chunks)
    return tuple(newchunks)


def rechunk_for_cohorts(
    array: DaskArray,
    axis: T_Axis,
    labels: np.ndarray,
    force_new_chunk_at: Sequence,
    chunksize: int | None = None,
    ignore_old_chunks: bool = False,
    debug: bool = False,
) -> DaskArray:
    """
    Rechunks array so that each new chunk contains groups that always occur together.

    Parameters
    ----------
    array : dask.array.Array
        array to rechunk
    axis : int
        Axis to rechunk
    labels : np.ndarray
        1D Group labels to align chunks with. This routine works
        well when ``labels`` has repeating patterns: e.g.
        ``1, 2, 3, 1, 2, 3, 4, 1, 2, 3`` though there is no requirement
        that the pattern must contain sequences.
    force_new_chunk_at : Sequence
        Labels at which we always start a new chunk. For
        the example ``labels`` array, this would be `1`.
    chunksize : int, optional
        nominal chunk size. Chunk size is exceeded when the label
        in ``force_new_chunk_at`` is less than ``chunksize//2`` elements away.
        If None, uses median chunksize along axis.

    Returns
    -------
    dask.array.Array
        rechunked array
    """
    if chunksize is None:
        chunksize = np.median(array.chunks[axis]).astype(int)

    if len(labels) != array.shape[axis]:
        raise ValueError(
            "labels must be equal to array.shape[axis]. "
            f"Received length {len(labels)}.  Expected length {array.shape[axis]}"
        )

    force_new_chunk_at = _atleast_1d(force_new_chunk_at)
    oldchunks = array.chunks[axis]
    oldbreaks = np.insert(np.cumsum(oldchunks), 0, 0)
    if debug:
        labels_at_breaks = labels[oldbreaks[:-1]]
        print(labels_at_breaks[:40])

    isbreak = np.isin(labels, force_new_chunk_at)
    if not np.any(isbreak):
        raise ValueError("One or more labels in ``force_new_chunk_at`` not present in ``labels``.")

    divisions = []
    counter = 1
    for idx, lab in enumerate(labels):
        if lab in force_new_chunk_at or idx == 0:
            divisions.append(idx)
            counter = 1
            continue

        next_break = np.nonzero(isbreak[idx:])[0]
        if next_break.any():
            next_break_is_close = next_break[0] <= chunksize // 2
        else:
            next_break_is_close = False

        if (not ignore_old_chunks and idx in oldbreaks) or (counter >= chunksize and not next_break_is_close):
            divisions.append(idx)
            counter = 1
            continue

        counter += 1

    divisions.append(len(labels))
    if debug:
        labels_at_breaks = labels[divisions[:-1]]
        print(labels_at_breaks[:40])

    newchunks = tuple(np.diff(divisions))
    if debug:
        print(divisions[:10], newchunks[:10])
        print(divisions[-10:], newchunks[-10:])
    assert sum(newchunks) == len(labels)

    if newchunks == array.chunks[axis]:
        return array
    else:
        return array.rechunk({axis: newchunks})


def rechunk_for_blockwise(
    array: DaskArray, axis: T_Axis, labels: np.ndarray, *, force: bool = True
) -> tuple[T_MethodOpt, DaskArray]:
    """
    Rechunks array so that group boundaries line up with chunk boundaries, allowing
    embarrassingly parallel group reductions.

    This only works when the groups are sequential
    (e.g. labels = ``[0,0,0,1,1,1,1,2,2]``).
    Such patterns occur when using ``.resample``.

    Parameters
    ----------
    array : DaskArray
        Array to rechunk
    axis : int
        Axis along which to rechunk the array.
    labels : np.ndarray
        Group labels

    Returns
    -------
    DaskArray
        Rechunked array
    """

    chunks = array.chunks[axis]
    if len(chunks) == 1:
        return "blockwise", array

    # import dask
    # from dask.utils import parse_bytes
    # factor = parse_bytes(dask.config.get("array.chunk-size")) / (
    #     math.prod(array.chunksize) * array.dtype.itemsize
    # )
    # if factor > BLOCKWISE_DEFAULT_ARRAY_CHUNK_SIZE_FACTOR:
    #     new_constant_chunks = math.ceil(factor) * max(chunks)
    #     q, r = divmod(array.shape[axis], new_constant_chunks)
    #     new_input_chunks = (new_constant_chunks,) * q + (r,)
    # else:
    new_input_chunks = chunks

    # FIXME: this should be unnecessary?
    labels = factorize_((labels,), axes=())[0]
    newchunks = _get_optimal_chunks_for_groups(new_input_chunks, labels)
    if newchunks == chunks:
        return "blockwise", array

    Δn = abs(len(newchunks) - len(new_input_chunks))
    if pass_num_chunks_threshold := (
        Δn / len(new_input_chunks) < OPTIONS["rechunk_blockwise_num_chunks_threshold"]
    ):
        logger.debug("blockwise rechunk passes num chunks threshold")
    if pass_chunk_size_threshold := (
        # we just pick the max because number of chunks may have changed.
        (abs(max(newchunks) - max(new_input_chunks)) / max(new_input_chunks))
        < OPTIONS["rechunk_blockwise_chunk_size_threshold"]
    ):
        logger.debug("blockwise rechunk passes chunk size change threshold")

    if force or (pass_num_chunks_threshold and pass_chunk_size_threshold):
        logger.debug("Rechunking to enable blockwise.")
        return "blockwise", array.rechunk({axis: newchunks})
    else:
        logger.debug("Didn't meet thresholds to do automatic rechunking for blockwise reductions.")
        return None, array


__all__ = [
    "rechunk_for_blockwise",
    "rechunk_for_cohorts",
]



"""Dask-specific functions for groupby operations."""

from __future__ import annotations

import itertools
import operator
from collections.abc import Callable, Sequence
from functools import partial
from numbers import Integral
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd
import toolz as tlz

if TYPE_CHECKING:
    from typing import Literal

    from .aggregations import Aggregation
    from .core import T_Axes, T_Engine, T_Method
    from .lib import ArrayLayer
    from .reindex import ReindexArrayType, ReindexStrategy
    from .types import DaskArray, Graph, IntermediateDict, T_By

    T_ScanMethod = Literal["blelloch", "blockwise"]

from .aggregations import Scan, scan_binary_op
from .core import (
    DUMMY_AXIS,
    _get_chunk_reduction,
    _reduce_blockwise,
    _unique,
    chunk_argreduce,
    chunk_reduce,
)
from .lib import _is_arg_reduction, _is_first_last_reduction, _issorted, identity
from .reindex import (
    ReindexArrayType,
    ReindexStrategy,
    reindex_,
)
from .scan import _finalize_scan, _zip, chunk_scan, grouped_reduce
from .types import FinalResultsDict, IntermediateDict
from .xrutils import is_duck_dask_array, notnull


def listify_groups(x: IntermediateDict):
    return list(np.atleast_1d(x["groups"].squeeze()))


def _find_unique_groups(x_chunk) -> np.ndarray:
    from dask.base import flatten
    from dask.utils import deepmap

    unique_groups = _unique(np.asarray(tuple(flatten(deepmap(listify_groups, x_chunk)))))
    unique_groups = unique_groups[notnull(unique_groups)]

    if len(unique_groups) == 0:
        unique_groups = np.array([np.nan])
    return unique_groups


def _conc2(x_chunk, key1, key2=slice(None), axis=None) -> np.ndarray:
    """copied from dask.array.reductions.mean_combine"""
    from dask.array.core import _concatenate2
    from dask.utils import deepmap

    mapped = deepmap(lambda x: x[key1][key2], x_chunk)
    return _concatenate2(mapped, axes=axis)


def reindex_intermediates(
    x: IntermediateDict, agg: Aggregation, unique_groups, array_type
) -> IntermediateDict:
    new_shape = x["groups"].shape[:-1] + (len(unique_groups),)
    newx: IntermediateDict = {"groups": np.broadcast_to(unique_groups, new_shape)}
    newx["intermediates"] = tuple(
        reindex_(
            v,
            from_=np.atleast_1d(x["groups"].squeeze()),
            to=pd.Index(unique_groups),
            fill_value=f,
            array_type=array_type,
        )
        for v, f in zip(x["intermediates"], agg.fill_value["intermediate"])
    )
    return newx


def _simple_combine(
    x_chunk,
    agg: Aggregation,
    axis,
    keepdims: bool,
    reindex: ReindexStrategy,
    is_aggregate: bool = False,
) -> IntermediateDict:
    """
    'Simple' combination of blockwise results.

    1. After the blockwise groupby-reduce, all blocks contain a value for all possible groups,
       and are of the same shape; i.e. reindex must have been True
    2. _expand_dims was used to insert an extra axis DUMMY_AXIS
    3. Here we concatenate along DUMMY_AXIS, and then call the combine function along
       DUMMY_AXIS
    4. At the final aggregate step, we squeeze out DUMMY_AXIS
    """
    import warnings

    from dask.array.core import deepfirst
    from dask.utils import deepmap

    if not reindex.blockwise:
        # We didn't reindex at the blockwise step
        # So now reindex before combining by reducing along DUMMY_AXIS
        unique_groups = _find_unique_groups(x_chunk)
        x_chunk = deepmap(
            partial(
                reindex_intermediates,
                agg=agg,
                unique_groups=unique_groups,
                array_type=reindex.array_type,
            ),
            x_chunk,
        )
    else:
        unique_groups = deepfirst(x_chunk)["groups"]

    results: IntermediateDict = {"groups": unique_groups}
    results["intermediates"] = []
    axis_ = axis[:-1] + (DUMMY_AXIS,)
    for idx, combine in enumerate(agg.simple_combine):
        array = _conc2(x_chunk, key1="intermediates", key2=idx, axis=axis_)
        assert array.ndim >= 2
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", r"All-NaN (slice|axis) encountered")
            assert callable(combine)
            result = combine(array, axis=axis_, keepdims=True)
        if is_aggregate:
            # squeeze out DUMMY_AXIS if this is the last step i.e. called from _aggregate
            # can't just pass DUMMY_AXIS, because of sparse.COO
            result = result.squeeze(range(result.ndim)[DUMMY_AXIS])
        results["intermediates"].append(result)
    return results


def _extract_result(result_dict: FinalResultsDict, key) -> np.ndarray:
    from dask.array.core import deepfirst

    # deepfirst should be not be needed here but sometimes we receive a list of dict?
    return deepfirst(result_dict)[key]


def _expand_dims(results: IntermediateDict) -> IntermediateDict:
    results["intermediates"] = tuple(
        np.expand_dims(array, axis=DUMMY_AXIS) for array in results["intermediates"]
    )
    return results


def _aggregate(
    x_chunk,
    combine,
    agg: Aggregation,
    expected_groups,
    axis,
    keepdims: bool,
    fill_value,
    reindex: ReindexStrategy,
) -> FinalResultsDict:
    """Final aggregation step of tree reduction"""
    from .core import _finalize_results

    results = combine(x_chunk, agg, axis, keepdims, is_aggregate=True)
    return _finalize_results(results, agg, axis, expected_groups, reindex=reindex)


def _unify_chunks(array, by):
    from dask.array import from_array, unify_chunks

    inds = tuple(range(array.ndim))

    # Unifying chunks is necessary for argreductions.
    # We need to rechunk before zipping up with the index
    # let's always do it anyway
    if not is_duck_dask_array(by):
        # chunk numpy arrays like the input array
        # This removes an extra rechunk-merge layer that would be
        # added otherwise
        chunks = tuple(array.chunks[ax] if by.shape[ax] != 1 else (1,) for ax in range(-by.ndim, 0))

        by = from_array(by, chunks=chunks)
    _, (array, by) = unify_chunks(array, inds, by, inds[-by.ndim :])

    return array, by


def _grouped_combine(
    x_chunk,
    agg: Aggregation,
    axis: T_Axes,
    keepdims: bool,
    engine: T_Engine,
    is_aggregate: bool = False,
    sort: bool = True,
) -> IntermediateDict:
    """Combine intermediates step of tree reduction."""
    from dask.utils import deepmap

    combine = agg.combine

    if isinstance(x_chunk, dict):
        # Only one block at final step; skip one extra groupby
        return x_chunk

    if len(axis) != 1:
        # when there's only a single axis of reduction, we can just concatenate later,
        # reindexing is unnecessary
        # I bet we can minimize the amount of reindexing for mD reductions too, but it's complicated
        unique_groups = _find_unique_groups(x_chunk)
        x_chunk = deepmap(
            partial(
                reindex_intermediates, agg=agg, unique_groups=unique_groups, array_type=ReindexArrayType.AUTO
            ),
            x_chunk,
        )

    # these are negative axis indices useful for concatenating the intermediates
    neg_axis = tuple(range(-len(axis), 0))

    groups = _conc2(x_chunk, "groups", axis=neg_axis)

    if agg.reduction_type == "argreduce":
        # If "nanlen" was added for masking later, we need to account for that
        if agg.chunk[-1] == "nanlen":
            slicer = slice(None, -1)
        else:
            slicer = slice(None, None)

        # We need to send the intermediate array values & indexes at the same time
        # intermediates are (value e.g. max, index e.g. argmax, counts)
        array_idx = tuple(_conc2(x_chunk, key1="intermediates", key2=idx, axis=axis) for idx in (0, 1))

        # for a single element along axis, we don't want to run the argreduction twice
        # This happens when we are reducing along an axis with a single chunk.
        avoid_reduction = array_idx[0].shape[axis[0]] == 1
        if avoid_reduction:
            results: IntermediateDict = {
                "groups": groups,
                "intermediates": list(array_idx),
            }
        else:
            results = chunk_argreduce(
                array_idx,
                groups,
                # count gets treated specially next
                func=combine[slicer],  # type: ignore[arg-type]
                axis=axis,
                expected_groups=None,
                fill_value=agg.fill_value["intermediate"][slicer],
                dtype=agg.dtype["intermediate"][slicer],
                engine=engine,
                sort=sort,
            )

        if agg.chunk[-1] == "nanlen":
            counts = _conc2(x_chunk, key1="intermediates", key2=2, axis=axis)

            if avoid_reduction:
                results["intermediates"].append(counts)
            else:
                # sum the counts
                results["intermediates"].append(
                    chunk_reduce(
                        counts,
                        groups,
                        func="sum",
                        axis=axis,
                        expected_groups=None,
                        fill_value=(0,),
                        dtype=(np.intp,),
                        engine=engine,
                        sort=sort,
                        user_dtype=agg.dtype["user"],
                    )["intermediates"][0]
                )

    elif agg.reduction_type == "reduce":
        # Here we reduce the intermediates individually
        results = {"groups": None, "intermediates": []}
        for idx, (combine_, fv, dtype) in enumerate(
            zip(combine, agg.fill_value["intermediate"], agg.dtype["intermediate"])
        ):
            assert combine_ is not None
            array = _conc2(x_chunk, key1="intermediates", key2=idx, axis=axis)
            if array.shape[-1] == 0:
                # No groups found in input data. Return to avoid a tree-reduce
                # step with no data.
                results["groups"] = groups
                results["intermediates"].append(array)
                continue
            reduced = chunk_reduce(
                array,
                groups,
                axis=axis,
                func=combine_,
                expected_groups=None,
                fill_value=(fv,),
                dtype=(dtype,),
                engine=engine,
                sort=sort,
                user_dtype=agg.dtype["user"],
            )
            # we had groups so this should've been set
            if results["groups"] is None:
                results["groups"] = reduced["groups"]
            results["intermediates"].append(reduced["intermediates"][0])

    # final pass and add keepdims=False.
    results["groups"] = results["groups"].squeeze() if not keepdims else results["groups"]

    return results


def dask_groupby_agg(
    array: DaskArray,
    by: T_By,
    *,
    agg: Aggregation,
    expected_groups: pd.RangeIndex | None,
    reindex: ReindexStrategy,
    axis: T_Axes = (),
    fill_value: Any = None,
    method: T_Method = "map-reduce",
    engine: T_Engine = "numpy",
    sort: bool = True,
    chunks_cohorts=None,
) -> tuple[DaskArray, tuple[pd.Index | np.ndarray | DaskArray]]:
    import dask.array
    from dask.array.core import slices_from_chunks
    from dask.highlevelgraph import HighLevelGraph

    from .dask_array_ops import _tree_reduce

    # I think _tree_reduce expects this
    assert isinstance(axis, Sequence)
    assert all(ax >= 0 for ax in axis)

    inds = tuple(range(array.ndim))
    name = f"groupby_{agg.name}"

    if expected_groups is None and reindex.blockwise:
        raise ValueError("reindex.blockwise must be False-y if expected_groups is not provided.")
    if method == "cohorts" and reindex.blockwise:
        raise ValueError("reindex.blockwise must be False-y if method is 'cohorts'.")

    by_input = by

    array, by = _unify_chunks(array, by)

    # tokenize here since by has already been hashed if its numpy
    token = dask.base.tokenize(array, by, agg, expected_groups, axis, method)

    # preprocess the array:
    #   - for argreductions, this zips the index together with the array block
    #   - not necessary for blockwise with argreductions
    #   - if this is needed later, we can fix this then
    if agg.preprocess and method != "blockwise":
        array = agg.preprocess(array, axis=axis)

    # 1. We first apply the groupby-reduction blockwise to generate "intermediates"
    # 2. These intermediate results are combined to generate the final result using a
    #    "map-reduce" or "tree reduction" approach.
    #    There are two ways:
    #    a. "_simple_combine": Where it makes sense, we tree-reduce the reduction,
    #        NOT the groupby-reduction for a speed boost. This is what xhistogram does (effectively),
    #        It requires that all blocks contain all groups after the initial blockwise step (1) i.e.
    #        reindex.blockwise=True, and we must know expected_groups
    #    b. "_grouped_combine": A more general solution where we tree-reduce the groupby reduction.
    #       This allows us to discover groups at compute time, support argreductions, lower intermediate
    #       memory usage (but method="cohorts" would also work to reduce memory in some cases)
    labels_are_unknown = is_duck_dask_array(by_input) and expected_groups is None
    do_grouped_combine = (
        _is_arg_reduction(agg)
        or labels_are_unknown
        or (_is_first_last_reduction(agg) and array.dtype.kind != "f")
    )
    do_simple_combine = not do_grouped_combine

    if method == "blockwise":
        #  use the "non dask" code path, but applied blockwise
        blockwise_method = partial(_reduce_blockwise, agg=agg, fill_value=fill_value, reindex=reindex)
    else:
        # choose `chunk_reduce` or `chunk_argreduce`
        blockwise_method = partial(
            _get_chunk_reduction(agg.reduction_type),
            func=agg.chunk,
            reindex=reindex.blockwise,
            fill_value=agg.fill_value["intermediate"],
            dtype=agg.dtype["intermediate"],
            user_dtype=agg.dtype["user"],
        )
        if do_simple_combine:
            # Add a dummy dimension that then gets reduced over
            blockwise_method = tlz.compose(_expand_dims, blockwise_method)

    # apply reduction on chunk
    intermediate = dask.array.blockwise(
        partial(
            blockwise_method,
            axis=axis,
            expected_groups=expected_groups if reindex.blockwise else None,
            engine=engine,
            sort=sort,
        ),
        # output indices are the same as input indices
        # Unlike xhistogram, we don't always know what the size of the group
        # dimension will be unless reindex=True
        inds,
        array,
        inds,
        by,
        inds[-by.ndim :],
        concatenate=False,
        dtype=array.dtype,  # this is purely for show
        meta=array._meta,
        align_arrays=False,
        name=f"{name}-chunk-{token}",
    )

    group_chunks: tuple[tuple[int | float, ...]]

    if method in ["map-reduce", "cohorts"]:
        combine: Callable[..., IntermediateDict] = (
            partial(_simple_combine, reindex=reindex)
            if do_simple_combine
            else partial(_grouped_combine, engine=engine, sort=sort)
        )

        tree_reduce = partial(
            dask.array.reductions._tree_reduce,
            name=f"{name}-simple-reduce",
            dtype=array.dtype,
            axis=axis,
            keepdims=True,
            concatenate=False,
        )
        aggregate = partial(_aggregate, combine=combine, agg=agg, fill_value=fill_value, reindex=reindex)

        # Each chunk of `reduced`` is really a dict mapping
        # 1. reduction name to array
        # 2. "groups" to an array of group labels
        # Note: it does not make sense to interpret axis relative to
        # shape of intermediate results after the blockwise call
        if method == "map-reduce":
            reduced = tree_reduce(
                intermediate,
                combine=partial(combine, agg=agg),
                aggregate=partial(aggregate, expected_groups=expected_groups),
            )
            if labels_are_unknown:
                groups = _extract_unknown_groups(reduced, dtype=by.dtype)
                group_chunks = ((np.nan,),)
            else:
                assert expected_groups is not None
                groups = (expected_groups,)
                group_chunks = ((len(expected_groups),),)

        elif method == "cohorts":
            assert chunks_cohorts
            block_shape = array.blocks.shape[-len(axis) :]

            out_name = f"{name}-reduce-{method}-{token}"
            groups_ = []
            chunks_as_array = tuple(np.array(c) for c in array.chunks)
            dsk: Graph = {}
            for icohort, (blks, cohort) in enumerate(chunks_cohorts.items()):
                cohort_index = pd.Index(cohort)
                reindexer = (
                    partial(
                        reindex_intermediates,
                        agg=agg,
                        unique_groups=cohort_index,
                        array_type=reindex.array_type,
                    )
                    if do_simple_combine
                    else identity
                )
                subset = subset_to_blocks(intermediate, blks, block_shape, reindexer, chunks_as_array)
                dsk |= subset.layer  # type: ignore[operator]
                # now that we have reindexed, we can set reindex=True explicitlly
                new_reindex = ReindexStrategy(blockwise=do_simple_combine, array_type=reindex.array_type)
                _tree_reduce(
                    subset,
                    out_dsk=dsk,
                    name=out_name,
                    block_index=icohort,
                    axis=axis,
                    combine=partial(combine, agg=agg, reindex=new_reindex, keepdims=True),
                    aggregate=partial(
                        aggregate, expected_groups=cohort_index, reindex=new_reindex, keepdims=True
                    ),
                )
                # This is done because pandas promotes to 64-bit types when an Index is created
                # So we use the index to generate the return value for consistency with "map-reduce"
                # This is important on windows
                groups_.append(cohort_index.values)

            graph = HighLevelGraph.from_collections(out_name, dsk, dependencies=[intermediate])

            out_chunks = list(array.chunks)
            out_chunks[axis[-1]] = tuple(len(c) for c in chunks_cohorts.values())
            for ax in axis[:-1]:
                out_chunks[ax] = (1,)
            reduced = dask.array.Array(graph, out_name, out_chunks, meta=array._meta)

            groups = (np.concatenate(groups_),)
            group_chunks = (tuple(len(cohort) for cohort in groups_),)

    elif method == "blockwise":
        reduced = intermediate
        if reindex.blockwise:
            if TYPE_CHECKING:
                assert expected_groups is not None
            # TODO: we could have `expected_groups` be a dask array with appropriate chunks
            # for now, we have a numpy array that is interpreted as listing all group labels
            # that are present in every chunk
            groups = (expected_groups,)
            group_chunks = ((len(expected_groups),),)
        else:
            # TODO: use chunks_cohorts here; hard because chunks_cohorts does not include all-NaN blocks
            #       but the array after applying the blockwise op; does. We'd have to insert a subsetting op.
            # Here one input chunk → one output chunks
            # find number of groups in each chunk, this is needed for output chunks
            # along the reduced axis
            # TODO: this logic is very specialized for the resampling case
            slices = slices_from_chunks(tuple(array.chunks[ax] for ax in axis))
            groups_in_block = tuple(_unique(by_input[slc]) for slc in slices)
            groups = (np.concatenate(groups_in_block),)
            ngroups_per_block = tuple(len(grp) for grp in groups_in_block)
            group_chunks = (ngroups_per_block,)
    else:
        raise ValueError(f"Unknown method={method}.")

    # Adjust output for any new dimensions added, example for multiple quantiles
    new_dims_shape = tuple(dim.size for dim in agg.new_dims if not dim.is_scalar)
    new_inds = tuple(range(-len(new_dims_shape), 0))
    out_inds = new_inds + inds[: -len(axis)] + (inds[-1],)
    output_chunks = new_dims_shape + reduced.chunks[: -len(axis)] + group_chunks
    new_axes = dict(zip(new_inds, new_dims_shape))

    if method == "blockwise" and len(axis) > 1:
        # The final results are available but the blocks along axes
        # need to be reshaped to axis=-1
        # I don't know that this is possible with blockwise
        # All other code paths benefit from an unmaterialized Blockwise layer
        reduced = _collapse_blocks_along_axes(reduced, axis, group_chunks)

    # Can't use map_blocks because it forces concatenate=True along drop_axes,
    result = dask.array.blockwise(
        _extract_result,
        out_inds,
        reduced,
        inds,
        adjust_chunks=dict(zip(out_inds, output_chunks)),
        key=agg.name,
        name=f"{name}-{token}",
        concatenate=False,
        new_axes=new_axes,
        meta=reindex.get_dask_meta(array, dtype=agg.dtype["final"], fill_value=agg.fill_value[agg.name]),
    )

    return (result, groups)


def dask_groupby_scan(array, by, axes: T_Axes, agg: Scan, method: T_ScanMethod = "blelloch") -> DaskArray:
    """Grouped scan for dask arrays.

    Parameters
    ----------
    array : DaskArray
        Input array to scan.
    by : DaskArray
        Group labels array, must have same chunks as array along scan axis.
    axes : T_Axes
        Tuple of axes to scan along (must be single axis).
    agg : Scan
        Scan aggregation specification.
    method : {"blelloch", "blockwise"}, optional
        Scan method to use:
        - "blelloch": Blelloch parallel prefix scan algorithm, allows scanning
          across chunk boundaries using tree reduction. Default.
        - "blockwise": Each chunk is processed independently. Only valid when
          all members of each group are contained within a single chunk.

    Returns
    -------
    DaskArray
        Result of the grouped scan with same shape and chunks as input.
    """
    from dask.array import map_blocks
    from dask.array.reductions import cumreduction as scan
    from dask.base import tokenize

    if len(axes) > 1:
        raise NotImplementedError("Scans are only supported along a single axis.")
    (axis,) = axes

    array, by = _unify_chunks(array, by)

    # Include method in token to differentiate task graphs
    token = tokenize(array, by, agg, axes, method)

    # 1. zip together group indices & array
    zipped = map_blocks(
        _zip,
        by,
        array,
        dtype=array.dtype,
        meta=array._meta,
        name=f"groupby-scan-preprocess-{token}",
    )

    # 2. Run the scan
    if method == "blockwise":
        # Apply chunk_scan blockwise - each block independently
        scan_func = partial(chunk_scan, agg=agg, axis=axis, dtype=agg.dtype)
        scanned = map_blocks(
            scan_func,
            zipped,
            dtype=agg.dtype,
            meta=array._meta,
            name=f"groupby-scan-{token}",
        )
    else:
        # Use Blelloch parallel prefix scan algorithm
        scan_ = partial(chunk_scan, agg=agg)
        # dask tokenizing error workaround
        scan_.__name__ = scan_.func.__name__  # type: ignore[attr-defined]

        scanned = scan(
            func=scan_,
            binop=partial(scan_binary_op, agg=agg),
            ident=agg.identity,
            x=zipped,
            axis=axis,
            # TODO: support method="sequential" here.
            method="blelloch",
            preop=partial(grouped_reduce, agg=agg),
            dtype=agg.dtype,
        )

    # 3. Extract final result
    result = map_blocks(
        partial(_finalize_scan, dtype=agg.dtype),
        scanned,
        dtype=agg.dtype,
        name=f"groupby-scan-finalize-{token}",
    )

    assert result.chunks == array.chunks

    return result


def _normalize_indexes(ndim: int, flatblocks: Sequence[int], blkshape: tuple[int, ...]) -> tuple:
    """
    .blocks accessor can only accept one iterable at a time,
    but can handle multiple slices.
    To minimize tasks and layers, we normalize to produce slices
    along as many axes as possible, and then repeatedly apply
    any remaining iterables in a loop.

    TODO: move this upstream
    """
    unraveled = np.unravel_index(flatblocks, blkshape)

    normalized: list[int | slice | list[int]] = []
    for ax, idx in enumerate(unraveled):
        i = _unique(idx).squeeze()
        if i.ndim == 0:
            normalized.append(i.item())
        else:
            if len(i) == blkshape[ax] and np.array_equal(i, np.arange(blkshape[ax])):
                normalized.append(slice(None))
            elif _issorted(i) and np.array_equal(i, np.arange(i[0], i[-1] + 1)):
                start = None if i[0] == 0 else i[0]
                stop = i[-1] + 1
                stop = None if stop == blkshape[ax] else stop
                normalized.append(slice(start, stop))
            else:
                normalized.append(list(i))
    full_normalized = (slice(None),) * (ndim - len(normalized)) + tuple(normalized)

    # has no iterables
    noiter = list(i if not hasattr(i, "__len__") else slice(None) for i in full_normalized)
    # has all iterables
    alliter = {ax: i for ax, i in enumerate(full_normalized) if hasattr(i, "__len__")}

    mesh = dict(zip(alliter.keys(), np.ix_(*alliter.values())))  # type: ignore[arg-type, var-annotated]

    full_tuple = tuple(i if ax not in mesh else mesh[ax] for ax, i in enumerate(noiter))

    return full_tuple


def subset_to_blocks(
    array: DaskArray,
    flatblocks: Sequence[int],
    blkshape: tuple[int, ...] | None = None,
    reindexer=identity,
    chunks_as_array: tuple[np.ndarray, ...] | None = None,
) -> ArrayLayer:
    """
    Advanced indexing of .blocks such that we always get a regular array back.

    Parameters
    ----------
    array : dask.array
    flatblocks : flat indices of blocks to extract
    blkshape : shape of blocks with which to unravel flatblocks

    Returns
    -------
    dask.array
    """
    from dask.base import tokenize

    from .lib import ArrayLayer

    if blkshape is None:
        blkshape = array.blocks.shape

    if chunks_as_array is None:
        chunks_as_array = tuple(np.array(c) for c in array.chunks)

    index = _normalize_indexes(array.ndim, flatblocks, blkshape)

    # These rest is copied from dask.array.core.py with slight modifications
    index = tuple(slice(k, k + 1) if isinstance(k, Integral) else k for k in index)

    name = "groupby-cohort-" + tokenize(array, index)
    new_keys = array._key_array[index]

    squeezed = tuple(np.squeeze(i) if isinstance(i, np.ndarray) else i for i in index)
    chunks = tuple(tuple(c[i].tolist()) for c, i in zip(chunks_as_array, squeezed))

    keys = itertools.product(*(range(len(c)) for c in chunks))
    layer: Graph = {(name,) + key: (reindexer, tuple(new_keys[key].tolist())) for key in keys}
    return ArrayLayer(layer=layer, chunks=chunks, name=name)


def _extract_unknown_groups(reduced, dtype) -> tuple[DaskArray]:
    import dask.array
    from dask.highlevelgraph import HighLevelGraph

    groups_token = f"group-{reduced.name}"
    first_block = reduced.ndim * (0,)
    layer: Graph = {(groups_token, 0): (operator.getitem, (reduced.name, *first_block), "groups")}
    groups: tuple[DaskArray] = (
        dask.array.Array(
            HighLevelGraph.from_collections(groups_token, layer, dependencies=[reduced]),
            groups_token,
            chunks=((np.nan,),),
            meta=np.array([], dtype=dtype),
        ),
    )

    return groups


def _collapse_blocks_along_axes(reduced: DaskArray, axis: T_Axes, group_chunks) -> DaskArray:
    import dask.array
    from dask.highlevelgraph import HighLevelGraph

    nblocks = tuple(reduced.numblocks[ax] for ax in axis)
    output_chunks = reduced.chunks[: -len(axis)] + ((1,) * (len(axis) - 1),) + group_chunks

    # extract results from the dict
    ochunks = tuple(range(len(chunks_v)) for chunks_v in output_chunks)
    layer2: dict[tuple, tuple] = {}
    name = f"reshape-{reduced.name}"

    for ochunk in itertools.product(*ochunks):
        inchunk = ochunk[: -len(axis)] + np.unravel_index(ochunk[-1], nblocks)
        layer2[(name, *ochunk)] = (reduced.name, *inchunk)

    layer2: Graph
    return dask.array.Array(
        HighLevelGraph.from_collections(name, layer2, dependencies=[reduced]),
        name,
        chunks=output_chunks,
        dtype=reduced.dtype,
    )


__all__ = [
    "_collapse_blocks_along_axes",
    "_extract_unknown_groups",
    "_grouped_combine",
    "_normalize_indexes",
    "_unify_chunks",
    "dask_groupby_agg",
    "dask_groupby_scan",
    "subset_to_blocks",
]



from collections import namedtuple
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, TypeAlias, TypedDict

import numpy as np

try:
    import cubed.Array as CubedArray
except ImportError:
    CubedArray = Any

try:
    import dask.array.Array as DaskArray
    from dask.typing import Graph
except ImportError:
    DaskArray = Any
    Graph: TypeAlias = Any  # type: ignore[no-redef,misc]


if TYPE_CHECKING:
    T_DuckArray: TypeAlias = np.ndarray | DaskArray | CubedArray
    T_By: TypeAlias = T_DuckArray
    T_Bys = tuple[T_By, ...]
    T_Axis = int
    T_Axes = tuple[T_Axis, ...]


IntermediateDict = dict[str | Callable, Any]
FinalResultsDict = dict[str, DaskArray | CubedArray | np.ndarray]


class FactorizeKwargs(TypedDict, total=False):
    """Used in _factorize_multiple"""

    by: "T_Bys"
    axes: "T_Axes"
    fastpath: bool
    reindex: bool
    sort: bool


FactorProps = namedtuple("FactorProps", "offset_group nan_sentinel nanmask")



from __future__ import annotations

import datetime
import logging
import math
from collections.abc import Callable, Sequence
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeAlias,
    TypeVar,
    overload,
)

import numpy as np
import pandas as pd

from . import xrdtypes
from .aggregate_flox import _prepare_for_flox
from .aggregations import (
    Aggregation,
    Scan,
    _atleast_1d,
    _initialize_aggregation,
    generic_aggregate,
    is_var_chunk_reduction,
    quantile_new_dims_func,
)
from .factorize import (
    _factorize_multiple,
    factorize_,
)
from .lib import sparse_array_type
from .rechunk import rechunk_for_blockwise
from .reindex import (
    ReindexArrayType,
    ReindexStrategy,
    reindex_,
)
from .xrutils import (
    _contains_cftime_datetimes,
    _to_pytimedelta,
    datetime_to_numeric,
    is_duck_array,
    is_duck_cubed_array,
    is_duck_dask_array,
    isnull,
    module_available,
    notnull,
)

if module_available("numpy", minversion="2.0.0"):
    from numpy.lib.array_utils import normalize_axis_tuple
else:
    from numpy.core.numeric import normalize_axis_tuple  # type: ignore[no-redef]

HAS_NUMBAGG = module_available("numbagg", minversion="0.3.0")
HAS_SPARSE = module_available("sparse")

if TYPE_CHECKING:
    from .types import CubedArray, DaskArray

    T_DuckArray: TypeAlias = np.ndarray | DaskArray | CubedArray  # Any ?
    T_By: TypeAlias = T_DuckArray
    T_Bys: TypeAlias = tuple[T_By, ...]
    T_ExpectIndex: TypeAlias = pd.Index
    T_ExpectIndexTuple: TypeAlias = tuple[T_ExpectIndex, ...]
    T_ExpectIndexOpt: TypeAlias = T_ExpectIndex | None
    T_ExpectIndexOptTuple: TypeAlias = tuple[T_ExpectIndexOpt, ...]
    T_Expect: TypeAlias = Sequence | np.ndarray | T_ExpectIndex
    T_ExpectTuple: TypeAlias = tuple[T_Expect, ...]
    T_ExpectOpt: TypeAlias = Sequence | np.ndarray | T_ExpectIndexOpt
    T_ExpectOptTuple: TypeAlias = tuple[T_ExpectOpt, ...]
    T_ExpectedGroups: TypeAlias = T_Expect | T_ExpectOptTuple
    T_ExpectedGroupsOpt: TypeAlias = T_ExpectedGroups | None
    T_Func: TypeAlias = str | Callable
    T_Funcs: TypeAlias = T_Func | Sequence[T_Func]
    T_Agg: TypeAlias = str | Aggregation
    T_Scan: TypeAlias = str | Scan
    T_Axis: TypeAlias = int
    T_Axes: TypeAlias = tuple[T_Axis, ...]
    T_AxesOpt: TypeAlias = T_Axis | T_Axes | None
    T_Dtypes: TypeAlias = np.typing.DTypeLike | Sequence[np.typing.DTypeLike] | None
    T_FillValues: TypeAlias = np.typing.ArrayLike | Sequence[np.typing.ArrayLike] | None
    T_Engine: TypeAlias = Literal["flox", "numpy", "numba", "numbagg"]
    T_EngineOpt: TypeAlias = None | T_Engine
    T_Method: TypeAlias = Literal["map-reduce", "blockwise", "cohorts"]
    T_MethodOpt: TypeAlias = None | Literal["map-reduce", "blockwise", "cohorts"]
    T_IsBins: TypeAlias = bool | Sequence[bool]

from .types import FinalResultsDict, IntermediateDict

T = TypeVar("T")

# This dummy axis is inserted using np.expand_dims
# and then reduced over during the combine stage by
# _simple_combine.
DUMMY_AXIS = -2


logger = logging.getLogger("flox")


from .cohorts import find_group_cohorts
from .lib import (
    _is_arg_reduction,
    _is_bool_supported_reduction,
    _is_first_last_reduction,
    _is_minmax_reduction,
    _is_nanlen,
    _is_reindex_sparse_supported_reduction,
    _issorted,
    _postprocess_numbagg,
    _should_auto_rechunk_blockwise,
)


def _get_expected_groups(by: T_By, sort: bool) -> T_ExpectIndex:
    if is_duck_dask_array(by):
        raise ValueError("Please provide expected_groups if not grouping by a numpy array.")
    flatby = by.reshape(-1)
    expected = pd.unique(flatby[notnull(flatby)])
    return _convert_expected_groups_to_index((expected,), isbin=(False,), sort=sort)[0]


def _get_chunk_reduction(reduction_type: Literal["reduce", "argreduce"]) -> Callable:
    if reduction_type == "reduce":
        return chunk_reduce
    elif reduction_type == "argreduce":
        return chunk_argreduce
    else:
        raise ValueError(f"Unknown reduction type: {reduction_type}")


def _move_reduce_dims_to_end(arr: np.ndarray, axis: T_Axes) -> np.ndarray:
    """Transpose `arr` by moving `axis` to the end."""
    axis = tuple(axis)
    order = tuple(ax for ax in np.arange(arr.ndim) if ax not in axis) + axis
    arr = arr.transpose(order)
    return arr


def _collapse_axis(arr: np.ndarray, naxis: int) -> np.ndarray:
    """Reshape so that the last `naxis` axes are collapsed to one axis."""
    newshape = arr.shape[:-naxis] + (math.prod(arr.shape[-naxis:]),)
    return arr.reshape(newshape)


def _unique(a: np.ndarray) -> np.ndarray:
    """Much faster to use pandas unique and sort the results.
    np.unique sorts before uniquifying and is slow."""
    return np.sort(pd.unique(a.reshape(-1)))


def chunk_argreduce(
    array_plus_idx: tuple[np.ndarray, ...],
    by: np.ndarray,
    func: T_Funcs,
    expected_groups: pd.Index | None,
    axis: T_AxesOpt,
    fill_value: T_FillValues,
    dtype: T_Dtypes = None,
    reindex: bool = False,
    engine: T_Engine = "numpy",
    sort: bool = True,
    user_dtype=None,
) -> IntermediateDict:
    """
    Per-chunk arg reduction.

    Expects a tuple of (array, index along reduction axis). Inspired by
    dask.array.reductions.argtopk
    """
    array, idx = array_plus_idx
    by = np.broadcast_to(by, array.shape)

    results = chunk_reduce(
        array,
        by,
        func,
        expected_groups=None,
        axis=axis,
        fill_value=fill_value,
        dtype=dtype,
        engine=engine,
        sort=sort,
        user_dtype=user_dtype,
    )
    if not all(isnull(results["groups"])):
        idx = np.broadcast_to(idx, array.shape)

        # array, by get flattened to 1D before passing to npg
        # so the indexes need to be unraveled
        newidx = np.unravel_index(results["intermediates"][1], array.shape)

        # Now index into the actual "global" indexes `idx`
        results["intermediates"][1] = idx[newidx]

    if reindex and expected_groups is not None:
        results["intermediates"][1] = reindex_(
            results["intermediates"][1],
            results["groups"].squeeze(),
            expected_groups,
            fill_value=0,
        )

    assert results["intermediates"][0].shape == results["intermediates"][1].shape

    return results


def chunk_reduce(
    array: np.ndarray,
    by: np.ndarray,
    func: T_Funcs,
    expected_groups: pd.Index | None,
    axis: T_AxesOpt = None,
    fill_value: T_FillValues = None,
    dtype: T_Dtypes = None,
    reindex: bool = False,
    engine: T_Engine = "numpy",
    kwargs: Sequence[dict] | None = None,
    sort: bool = True,
    user_dtype=None,
) -> IntermediateDict:
    """
    Wrapper for numpy_groupies aggregate that supports nD ``array`` and
    mD ``by``.

    Core groupby reduction using numpy_groupies. Uses ``pandas.factorize`` to factorize
    ``by``. Offsets the groups if not reducing along all dimensions of ``by``.
    Always ravels ``by`` to 1D, flattens appropriate dimensions of array.

    When dask arrays are passed to groupby_reduce, this function is called on every
    block.

    Parameters
    ----------
    array : numpy.ndarray
        Array of values to reduced
    by : numpy.ndarray
        Array to group by.
    func : str or Callable or Sequence[str] or Sequence[Callable]
        Name of reduction or function, passed to numpy_groupies.
        Supports multiple reductions.
    axis : (optional) int or Sequence[int]
        If None, reduce along all dimensions of array.
        Else reduce along specified axes.

    Returns
    -------
    dict
    """

    funcs = _atleast_1d(func)
    nfuncs = len(funcs)
    dtypes = _atleast_1d(dtype, nfuncs)
    fill_values = _atleast_1d(fill_value, nfuncs)
    kwargss = _atleast_1d({}, nfuncs) if kwargs is None else kwargs

    if isinstance(axis, Sequence):
        axes: T_Axes = axis
        nax = len(axes)
    else:
        nax = by.ndim
        axes = () if axis is None else (axis,) * nax

    assert by.ndim <= array.ndim

    final_array_shape = array.shape[:-nax] + (1,) * (nax - 1)
    final_groups_shape = (1,) * (nax - 1)

    if 1 < nax < by.ndim:
        # when axis is a tuple
        # collapse and move reduction dimensions to the end
        by = _collapse_axis(by, nax)
        array = _collapse_axis(array, nax)
        axes = (-1,)
        nax = 1

    # if indices=[2,2,2], npg assumes groups are (0, 1, 2);
    # and will return a result that is bigger than necessary
    # avoid by factorizing again so indices=[2,2,2] is changed to
    # indices=[0,0,0]. This is necessary when combining block results
    # factorize can handle strings etc unlike digitize
    group_idx, grps, found_groups_shape, _, size, props = factorize_(
        (by,), axes, expected_groups=(expected_groups,), reindex=bool(reindex), sort=sort
    )
    (groups,) = grps

    # do this *before* possible broadcasting below.
    # factorize_ has already taken care of offsetting
    if engine == "numbagg":
        seen_groups = _unique(group_idx)

    order = "C"
    if nax > 1:
        needs_broadcast = any(
            group_idx.shape[ax] != array.shape[ax] and group_idx.shape[ax] == 1 for ax in range(-nax, 0)
        )
        if needs_broadcast:
            # This is the dim=... case, it's a lot faster to ravel group_idx
            # in fortran order since group_idx is then sorted
            # I'm seeing 400ms -> 23ms for engine="flox"
            # Of course we are slower to ravel `array` but we avoid argsorting
            # both `array` *and* `group_idx` in _prepare_for_flox
            group_idx = np.broadcast_to(group_idx, array.shape[-by.ndim :])
            if engine == "flox":
                group_idx = group_idx.reshape(-1, order="F")
                order = "F"
    # always reshape to 1D along group dimensions
    newshape = array.shape[: array.ndim - by.ndim] + (math.prod(array.shape[-by.ndim :]),)
    array = array.reshape(newshape, order=order)  # type: ignore[call-overload]
    group_idx = group_idx.reshape(-1)

    assert group_idx.ndim == 1

    empty = np.all(props.nanmask)
    hasnan = np.any(props.nanmask)

    results: IntermediateDict = {"groups": [], "intermediates": []}
    if reindex and expected_groups is not None:
        # TODO: what happens with binning here?
        results["groups"] = expected_groups
    else:
        if empty:
            results["groups"] = np.array([np.nan])
        else:
            results["groups"] = groups

    # npg's argmax ensures that index of first "max" is returned assuming there
    # are many elements equal to the "max". Sorting messes this up totally.
    # so we skip this for argreductions
    if engine == "flox":
        # is_arg_reduction = any("arg" in f for f in func if isinstance(f, str))
        # if not is_arg_reduction:
        group_idx, array, _ = _prepare_for_flox(group_idx, array)

    final_array_shape += results["groups"].shape
    final_groups_shape += results["groups"].shape

    # we commonly have func=(..., "nanlen", "nanlen") when
    # counts are needed for the final result as well as for masking
    # optimize that out.
    previous_reduction: T_Func = ""
    for reduction, fv, kw, dt in zip(funcs, fill_values, kwargss, dtypes):
        # UGLY! but this is because the `var` breaks our design assumptions
        if empty and not is_var_chunk_reduction(reduction):
            result = np.full(shape=final_array_shape, fill_value=fv, like=array)
        elif _is_nanlen(reduction) and _is_nanlen(previous_reduction):
            result = results["intermediates"][-1]
        else:
            # fill_value here is necessary when reducing with "offset" groups
            kw_func = dict(size=size, dtype=dt, fill_value=fv)
            kw_func.update(kw)

            # UGLY! but this is because the `var` breaks our design assumptions
            if is_var_chunk_reduction(reduction):
                kw_func.update(engine=engine)

            if callable(reduction):
                # passing a custom reduction for npg to apply per-group is really slow!
                # So this `reduction` has to do the groupby-aggregation
                result = reduction(group_idx, array, **kw_func)
            else:
                result = generic_aggregate(
                    group_idx, array, axis=-1, engine=engine, func=reduction, **kw_func
                ).astype(dt, copy=False)
            if engine == "numbagg":
                result = _postprocess_numbagg(
                    result,
                    func=reduction,
                    size=size,
                    fill_value=fv,
                    # Unfortunately, we cannot reuse found_groups, it has not
                    # been "offset" and is really expected_groups in nearly all cases
                    seen_groups=seen_groups,
                )
            if hasnan:
                # remove NaN group label which should be last
                result = result[..., :-1]
            # TODO: Figure out how to generalize this
            if reduction in ("quantile", "nanquantile"):
                new_dims_shape = tuple(dim.size for dim in quantile_new_dims_func(**kw) if not dim.is_scalar)
            else:
                new_dims_shape = tuple()
            result = result.reshape(new_dims_shape + final_array_shape[:-1] + found_groups_shape)
        results["intermediates"].append(result)
        previous_reduction = reduction

    results["groups"] = np.broadcast_to(results["groups"], final_groups_shape)
    return results


def _squeeze_results(results: IntermediateDict, axis: T_Axes) -> IntermediateDict:
    # at the end we squeeze out extra dims
    groups = results["groups"]
    newresults: IntermediateDict = {"groups": [], "intermediates": []}
    newresults["groups"] = np.squeeze(
        groups, axis=tuple(ax for ax in range(groups.ndim - 1) if groups.shape[ax] == 1)
    )
    for v in results["intermediates"]:
        squeeze_ax = tuple(ax for ax in sorted(axis)[:-1] if v.shape[ax] == 1)
        newresults["intermediates"].append(np.squeeze(v, axis=squeeze_ax) if squeeze_ax else v)
    return newresults


def _finalize_results(
    results: IntermediateDict,
    agg: Aggregation,
    axis: T_Axes,
    expected_groups: pd.Index | None,
    reindex: ReindexStrategy,
) -> FinalResultsDict:
    """Finalize results by
    1. Squeezing out dummy dimensions
    2. Calling agg.finalize with intermediate results
    3. Mask using counts and fill with user-provided fill_value.
    4. reindex to expected_groups
    """
    squeezed = _squeeze_results(results, tuple(agg.num_new_vector_dims + ax for ax in axis))

    min_count = agg.min_count
    if min_count > 0:
        counts = squeezed["intermediates"][-1]
        squeezed["intermediates"] = squeezed["intermediates"][:-1]

    # finalize step
    finalized: FinalResultsDict = {}
    if agg.finalize is None:
        finalized[agg.name] = squeezed["intermediates"][0]
    else:
        finalized[agg.name] = agg.finalize(*squeezed["intermediates"], **agg.finalize_kwargs)

    fill_value = agg.fill_value["user"]
    if min_count > 0:
        count_mask = counts < min_count
        if count_mask.any() or reindex.array_type is ReindexArrayType.SPARSE_COO:
            # For one count_mask.any() prevents promoting bool to dtype(fill_value) unless
            # necessary
            if fill_value is None:
                raise ValueError("Filling is required but fill_value is None.")
            # This allows us to match xarray's type promotion rules
            # Use '==' instead of 'is', as Dask serialization can break identity checks.
            if fill_value == xrdtypes.NA:
                new_dtype, fill_value = xrdtypes.maybe_promote(finalized[agg.name].dtype)
                finalized[agg.name] = finalized[agg.name].astype(new_dtype)

            if isinstance(count_mask, sparse_array_type):
                # This is guaranteed because for `counts`, we use `fill_value=0`
                # and we are in the counts < (min_count > 0) branch.
                # We need to flip the fill_value so that the `np.where` call preserves
                # the fill_value of finalized[agg.name], this is needed when dask concatenates results.
                assert count_mask.fill_value is np.True_, "Logic bug. I expected fill_value to be True"
                count_mask = type(count_mask).from_numpy(count_mask.todense(), fill_value=False)

            finalized[agg.name] = np.where(count_mask, fill_value, finalized[agg.name])

    # Final reindexing has to be here to be lazy
    if not reindex.blockwise and expected_groups is not None:
        finalized[agg.name] = reindex_(
            finalized[agg.name],
            squeezed["groups"],
            expected_groups,
            fill_value=fill_value,
            array_type=reindex.array_type,
        )
        finalized["groups"] = expected_groups
    else:
        finalized["groups"] = squeezed["groups"]

    finalized[agg.name] = finalized[agg.name].astype(agg.dtype["final"], copy=False)
    return finalized


def _reduce_blockwise(
    array,
    by,
    agg: Aggregation,
    *,
    axis: T_Axes,
    expected_groups,
    fill_value: Any,
    engine: T_Engine,
    sort: bool,
    reindex: ReindexStrategy,
) -> FinalResultsDict:
    """
    Blockwise groupby reduction that produces the final result. This code path is
    also used for non-dask array aggregations.
    """
    # for pure numpy grouping, we just use npg directly and avoid "finalizing"
    # (agg.finalize = None). We still need to do the reindexing step in finalize
    # so that everything matches the dask version.
    agg.finalize = None

    assert agg.finalize_kwargs is not None
    finalize_kwargs_: tuple[dict[Any, Any], ...] = (agg.finalize_kwargs,) + ({},) + ({},)

    results = chunk_reduce(
        array,
        by,
        func=agg.numpy,
        axis=axis,
        expected_groups=expected_groups,
        # This fill_value should only apply to groups that only contain NaN observations
        # BUT there is funkiness when axis is a subset of all possible values
        # (see below)
        fill_value=agg.fill_value["numpy"],
        dtype=agg.dtype["numpy"],
        kwargs=finalize_kwargs_,
        engine=engine,
        sort=sort,
        reindex=bool(reindex.blockwise),
        user_dtype=agg.dtype["user"],
    )

    if _is_arg_reduction(agg):
        results["intermediates"][0] = np.unravel_index(results["intermediates"][0], array.shape)[-1]

    result = _finalize_results(results, agg, axis, expected_groups, reindex=reindex)
    return result


def _validate_reindex(
    reindex: ReindexStrategy | bool | None,
    func,
    method: T_MethodOpt,
    expected_groups,
    any_by_dask: bool,
    is_dask_array: bool,
    array_dtype: Any,
) -> ReindexStrategy:
    # logger.debug("Entering _validate_reindex: reindex is {}".format(reindex))  # noqa
    def first_or_last():
        return func in ["first", "last"] or (_is_first_last_reduction(func) and array_dtype.kind != "f")

    all_eager = not is_dask_array and not any_by_dask
    if reindex is True and not all_eager:
        if _is_arg_reduction(func):
            raise NotImplementedError
        if method == "cohorts" or (method == "blockwise" and not any_by_dask):
            raise ValueError("reindex=True is not a valid choice for method='blockwise' or method='cohorts'.")
        if first_or_last():
            raise ValueError("reindex must be None or False when func is 'first' or 'last.")

    if isinstance(reindex, ReindexStrategy):
        reindex_ = reindex
    else:
        reindex_ = ReindexStrategy(blockwise=reindex)

    if reindex_.blockwise is None:
        if method is None:
            # logger.debug("Leaving _validate_reindex: method = None, returning None")
            return ReindexStrategy(blockwise=None)

        if all_eager:
            return ReindexStrategy(blockwise=True)

        if first_or_last():
            # have to do the grouped_combine since there's no good fill_value
            # Also needed for nanfirst, nanlast with no-NaN dtypes
            return ReindexStrategy(blockwise=False)

        if method == "blockwise":
            # for grouping by dask arrays, we set reindex=True
            reindex_ = ReindexStrategy(blockwise=any_by_dask)

        elif _is_arg_reduction(func):
            reindex_ = ReindexStrategy(blockwise=False)

        elif method == "cohorts":
            reindex_ = ReindexStrategy(blockwise=False)

        elif method == "map-reduce":
            if expected_groups is None and any_by_dask:
                reindex_ = ReindexStrategy(blockwise=False)
            else:
                reindex_ = ReindexStrategy(blockwise=True)

    assert isinstance(reindex_, ReindexStrategy)
    # logger.debug("Leaving _validate_reindex: reindex is {}".format(reindex))  # noqa

    return reindex_


def _assert_by_is_aligned(shape: tuple[int, ...], by: T_Bys) -> None:
    assert all(b.ndim == by[0].ndim for b in by[1:])
    for idx, b in enumerate(by):
        if not all(j in [i, 1] for i, j in zip(shape[-b.ndim :], b.shape)):
            raise ValueError(
                "`array` and `by` arrays must be 'aligned' "
                "so that such that by_ is broadcastable to array.shape[-by.ndim:] "
                "for every array `by_` in `by`. "
                "Either array.shape[-by_.ndim :] == by_.shape or the only differences "
                "should be size-1 dimensions in by_."
                f"Received array of shape {shape} but "
                f"array {idx} in `by` has shape {b.shape}."
            )


@overload
def _convert_expected_groups_to_index(
    expected_groups: tuple[None, ...], isbin: Sequence[bool], sort: bool
) -> tuple[None, ...]: ...


@overload
def _convert_expected_groups_to_index(
    expected_groups: T_ExpectTuple, isbin: Sequence[bool], sort: bool
) -> T_ExpectIndexTuple: ...


def _convert_expected_groups_to_index(
    expected_groups: T_ExpectOptTuple, isbin: Sequence[bool], sort: bool
) -> T_ExpectIndexOptTuple:
    out: list[T_ExpectIndexOpt] = []
    for ex, isbin_ in zip(expected_groups, isbin):
        if isinstance(ex, pd.IntervalIndex) or (isinstance(ex, pd.Index) and not isbin_):
            if sort:
                out.append(ex.sort_values())
            else:
                out.append(ex)
        elif ex is not None:
            if isbin_:
                out.append(pd.IntervalIndex.from_breaks(ex))
            else:
                if sort:
                    ex = np.sort(ex)
                out.append(pd.Index(ex))
        else:
            assert ex is None
            out.append(None)
    return tuple(out)


@overload
def _validate_expected_groups(nby: int, expected_groups: None) -> tuple[None, ...]: ...


@overload
def _validate_expected_groups(nby: int, expected_groups: T_ExpectedGroups) -> T_ExpectTuple: ...


def _validate_expected_groups(nby: int, expected_groups: T_ExpectedGroupsOpt) -> T_ExpectOptTuple:
    if expected_groups is None:
        return (None,) * nby

    if nby == 1 and not isinstance(expected_groups, tuple):
        if isinstance(expected_groups, pd.Index | np.ndarray):
            return (expected_groups,)
        else:
            array = np.asarray(expected_groups)
            if np.issubdtype(array.dtype, np.integer):
                # preserve default dtypes
                # on pandas 1.5/2, on windows
                # when a list is passed
                array = array.astype(np.int64)
            return (array,)

    if nby > 1 and not isinstance(expected_groups, tuple):  # TODO: test for list
        raise ValueError(
            "When grouping by multiple variables, expected_groups must be a tuple "
            "of either arrays or objects convertible to an array (like lists). "
            "For example `expected_groups=(np.array([1, 2, 3]), ['a', 'b', 'c'])`."
            f"Received a {type(expected_groups).__name__} instead. "
            "When grouping by a single variable, you can pass an array or something "
            "convertible to an array for convenience: `expected_groups=['a', 'b', 'c']`."
        )

    if TYPE_CHECKING:
        assert isinstance(expected_groups, tuple)

    if len(expected_groups) != nby:
        raise ValueError(
            f"Must have same number of `expected_groups` (received {len(expected_groups)}) "
            f" and variables to group by (received {nby})."
        )

    return expected_groups


def _choose_method(
    method: T_MethodOpt, preferred_method: T_Method, agg: Aggregation, by, nax: int
) -> T_Method:
    if method is None:
        logger.debug("_choose_method: method is None")
        if agg.chunk == (None,):
            if preferred_method != "blockwise":
                raise ValueError(
                    f"Aggregation {agg.name} is only supported for `method='blockwise'`, "
                    "but the chunking is not right."
                )
            logger.debug("_choose_method: choosing 'blockwise'")
            return "blockwise"

        if nax != by.ndim:
            logger.debug("_choose_method: choosing 'map-reduce'")
            return "map-reduce"

        if _is_arg_reduction(agg) and preferred_method == "blockwise":
            return "cohorts"

        logger.debug(f"_choose_method: choosing preferred_method={preferred_method}")  # noqa
        return preferred_method
    else:
        return method


def _choose_engine(by, agg: Aggregation):
    dtype = agg.dtype["user"]

    not_arg_reduce = not _is_arg_reduction(agg)

    if agg.name in ["quantile", "nanquantile", "median", "nanmedian"]:
        logger.debug(f"_choose_engine: Choosing 'flox' since {agg.name}")
        return "flox"

    # numbagg only supports nan-skipping reductions
    # without dtype specified
    has_blockwise_nan_skipping = (agg.chunk[0] is None and "nan" in agg.name) or any(
        (isinstance(func, str) and "nan" in func) for func in agg.chunk
    )
    if HAS_NUMBAGG:
        if agg.name in ["all", "any"] or (not_arg_reduce and has_blockwise_nan_skipping and dtype is None):
            logger.debug("_choose_engine: Choosing 'numbagg'")
            return "numbagg"

    if not_arg_reduce and (not is_duck_dask_array(by) and _issorted(by)):
        logger.debug("_choose_engine: Choosing 'flox'")
        return "flox"
    else:
        logger.debug("_choose_engine: Choosing 'numpy'")
        return "numpy"


def groupby_reduce(
    array: np.ndarray | DaskArray,
    *by: T_By,
    func: T_Agg,
    expected_groups: T_ExpectedGroupsOpt = None,
    sort: bool = True,
    isbin: T_IsBins = False,
    axis: T_AxesOpt = None,
    fill_value=None,
    dtype: np.typing.DTypeLike = None,
    min_count: int | None = None,
    method: T_MethodOpt = None,
    engine: T_EngineOpt = None,
    reindex: ReindexStrategy | bool | None = None,
    finalize_kwargs: dict[Any, Any] | None = None,
) -> tuple[DaskArray, *tuple[np.ndarray | DaskArray, ...]]:
    """
    GroupBy reductions using tree reductions for dask.array

    Parameters
    ----------
    array : ndarray or DaskArray
        Array to be reduced, possibly nD
    *by : ndarray or DaskArray
        Array of labels to group over. Must be aligned with ``array`` so that
        ``array.shape[-by.ndim :] == by.shape`` or any disagreements in that
        equality check are for dimensions of size 1 in ``by``.
    func : {"all", "any", "count", "sum", "nansum", "mean", "nanmean", \
            "max", "nanmax", "min", "nanmin", "argmax", "nanargmax", "argmin", "nanargmin", \
            "quantile", "nanquantile", "median", "nanmedian", "mode", "nanmode", \
            "first", "nanfirst", "last", "nanlast"} or Aggregation
        Single function name or an Aggregation instance
    expected_groups : (optional) Sequence
        Expected unique labels.
    isbin : bool, optional
        Are ``expected_groups`` bin edges?
    sort : bool, optional
        Whether groups should be returned in sorted order. Only applies for dask
        reductions when ``method`` is not ``"map-reduce"``. For ``"map-reduce"``, the groups
        are always sorted.
    axis : None or int or Sequence[int], optional
        If None, reduce across all dimensions of ``by``,
        else reduce across corresponding axes of array.
        Negative integers are normalized using ``array.ndim``.
    fill_value : Any
        Value to assign when a label in ``expected_groups`` is not present.
    dtype : data-type , optional
        DType for the output. Can be anything that is accepted by ``np.dtype``.
    min_count : int, default: None
        The required number of valid values to perform the operation. If
        fewer than ``min_count`` non-NA values are present the result will be
        NA. Only used if ``skipna`` is set to True or defaults to True for the
        array's dtype.
    method : {"map-reduce", "blockwise", "cohorts"}, optional
        Note that this arg is chosen by default using heuristics.
        Strategy for reduction of dask arrays only.
          * ``"map-reduce"``:
            First apply the reduction blockwise on ``array``, then
            combine a few newighbouring blocks, apply the reduction.
            Continue until finalizing. Usually, ``func`` will need
            to be an ``Aggregation`` instance for this method to work.
            Common aggregations are implemented.
          * ``"blockwise"``:
            Only reduce using blockwise and avoid aggregating blocks
            together. Useful for resampling-style reductions where group
            members are always together. If  ``by`` is 1D,  ``array`` is automatically
            rechunked so that chunk boundaries line up with group boundaries
            i.e. each block contains all members of any group present
            in that block. For nD ``by``, you must make sure that all members of a group
            are present in a single block.
          * ``"cohorts"``:
            Finds group labels that tend to occur together ("cohorts"),
            indexes out cohorts and reduces that subset using "map-reduce",
            repeat for all cohorts. This works well for many time groupings
            where the group labels repeat at regular intervals like 'hour',
            'month', dayofyear' etc. Optimize chunking ``array`` for this
            method by first rechunking using ``rechunk_for_cohorts``
            (for 1D ``by`` only).
    engine : {"flox", "numpy", "numba", "numbagg"}, optional
        Algorithm to compute the groupby reduction on non-dask arrays and on each dask chunk:
          * ``"numpy"``:
            Use the vectorized implementations in ``numpy_groupies.aggregate_numpy``.
            This is the default choice because it works for most array types.
          * ``"flox"``:
            Use an internal implementation where the data is sorted so that
            all members of a group occur sequentially, and then numpy.ufunc.reduceat
            is to used for the reduction. This will fall back to ``numpy_groupies.aggregate_numpy``
            for a reduction that is not yet implemented.
          * ``"numba"``:
            Use the implementations in ``numpy_groupies.aggregate_numba``.
          * ``"numbagg"``:
            Use the reductions supported by ``numbagg.grouped``. This will fall back to ``numpy_groupies.aggregate_numpy``
            for a reduction that is not yet implemented.
    reindex : ReindexStrategy | bool, optional
        Whether to "reindex" the blockwise reduced results to ``expected_groups`` (possibly automatically detected).
        If True, the intermediate result of the blockwise groupby-reduction has a value for all expected groups,
        and the final result is a simple reduction of those intermediates. In nearly all cases, this is a significant
        boost in computation speed. For cases like time grouping, this may result in large intermediates relative to the
        original block size. Avoid that by using ``method="cohorts"``. By default, it is turned off for argreductions.
        By default, the type of ``array`` is preserved. You may optionally reindex to a sparse array type to further control memory
        in the case of ``expected_groups`` being very large. Pass a ``ReindexStrategy`` instance with the appropriate ``array_type``,
        for example (``reindex=ReindexStrategy(blockwise=False, array_type=ReindexArrayType.SPARSE_COO)``).
    finalize_kwargs : dict, optional
        Kwargs passed to finalize the reduction such as ``ddof`` for var, std or ``q`` for quantile.

    Returns
    -------
    result
        Aggregated result
    *groups
        Group labels

    See Also
    --------
    xarray.xarray_reduce
    """

    if engine == "flox" and _is_arg_reduction(func):
        raise NotImplementedError(
            "argreductions not supported for engine='flox' yet. Try engine='numpy' or engine='numba' instead."
        )

    if engine == "numbagg" and dtype is not None:
        raise NotImplementedError(
            "numbagg does not support the `dtype` kwarg. Either cast your "
            "input arguments to `dtype` or use a different `engine`: "
            "'flox' or 'numpy' or 'numba'. "
            "See https://github.com/numbagg/numbagg/issues/121."
        )

    if func in ["quantile", "nanquantile"]:
        if finalize_kwargs is None or "q" not in finalize_kwargs:
            raise ValueError("Please pass `q` for quantile calculations.")
        else:
            nq = len(_atleast_1d(finalize_kwargs["q"]))
            if nq > 1 and engine == "numpy":
                raise ValueError(
                    "Multiple quantiles not supported with engine='numpy'."
                    "Use engine='flox' instead (it is also much faster), "
                    "or set engine=None to use the default."
                )

    bys: T_Bys = tuple(np.asarray(b) if not is_duck_array(b) else b for b in by)
    nby = len(bys)
    by_is_dask = tuple(is_duck_dask_array(b) for b in bys)
    any_by_dask = any(by_is_dask)
    provided_expected = expected_groups is not None

    if engine == "numbagg" and _is_arg_reduction(func) and (any_by_dask or is_duck_dask_array(array)):
        # There is only one test that fails, but I can't figure
        # out why without deep debugging.
        # just disable for now.
        # test_groupby_reduce_axis_subset_against_numpy
        # for array is 3D dask, by is 3D dask, axis=2
        # We are falling back to numpy for the arg reduction,
        # so presumably something is going wrong
        raise NotImplementedError(
            "argreductions not supported for engine='numbagg' yet."
            "Try engine='numpy' or engine='numba' instead."
        )

    if method == "cohorts" and any_by_dask:
        raise ValueError(f"method={method!r} can only be used when grouping by numpy arrays.")

    if not is_duck_array(array):
        array = np.asarray(array)

    reindex = _validate_reindex(
        reindex,
        func,
        method,
        expected_groups,
        any_by_dask,
        is_duck_dask_array(array),
        array.dtype,
    )

    is_bool_array = np.issubdtype(array.dtype, bool) and not _is_bool_supported_reduction(func)
    array = array.astype(np.int_) if is_bool_array else array

    isbins = _atleast_1d(isbin, nby)

    _assert_by_is_aligned(array.shape, bys)

    expected_groups = _validate_expected_groups(nby, expected_groups)

    for idx, (expect, is_dask) in enumerate(zip(expected_groups, by_is_dask)):
        if is_dask and (reindex.blockwise or nby > 1) and expect is None:
            raise ValueError(
                f"`expected_groups` for array {idx} in `by` cannot be None since it is a dask.array."
            )

    # We convert to pd.Index since that lets us know if we are binning or not
    # (pd.IntervalIndex or not)
    expected_groups = _convert_expected_groups_to_index(expected_groups, isbins, sort)

    # Don't factorize early only when
    # grouping by dask arrays, and not having expected_groups
    factorize_early = not (
        # can't do it if we are grouping by dask array but don't have expected_groups
        any(is_dask and ex_ is None for is_dask, ex_ in zip(by_is_dask, expected_groups))
    )
    expected_: pd.RangeIndex | None
    if factorize_early:
        bys, final_groups, grp_shape = _factorize_multiple(
            bys,
            expected_groups,
            any_by_dask=any_by_dask,
            sort=sort,
        )
        expected_ = pd.RangeIndex(math.prod(grp_shape))
    else:
        assert expected_groups == (None,)
        expected_ = None

    assert len(bys) == 1
    (by_,) = bys

    if axis is None:
        axis_ = tuple(array.ndim + np.arange(-by_.ndim, 0))
    else:
        axis_ = normalize_axis_tuple(axis, array.ndim)
    nax = len(axis_)

    has_dask = is_duck_dask_array(array) or is_duck_dask_array(by_)
    has_cubed = is_duck_cubed_array(array) or is_duck_cubed_array(by_)

    if _should_auto_rechunk_blockwise(method, array, any_by_dask, by_):
        # Let's try rechunking for sorted 1D by.
        (single_axis,) = axis_
        method, array = rechunk_for_blockwise(array, single_axis, by_, force=False)

    is_first_last = _is_first_last_reduction(func)
    if is_first_last:
        if has_dask and nax != 1:
            raise ValueError(
                "For dask arrays: first, last, nanfirst, nanlast reductions are "
                "only supported along a single axis. Please reshape appropriately."
            )

        elif nax not in [1, by_.ndim]:
            raise ValueError(
                "first, last, nanfirst, nanlast reductions are only supported "
                "along a single axis or when reducing across all dimensions of `by`."
            )

    is_npdatetime = array.dtype.kind in "Mm"
    is_cftime = _contains_cftime_datetimes(array)
    requires_numeric = (
        (func not in ["count", "any", "all"] and not is_first_last)
        # Flox's count works with non-numeric and its faster than converting.
        or (func == "count" and engine != "flox")
        # TODO: needed for npg, move to aggregate_npg
        or (is_first_last and is_cftime)
    )
    if requires_numeric:
        if is_npdatetime:
            datetime_dtype = array.dtype
            array = array.view(np.int64)
        elif is_cftime:
            offset = array.min()
            assert offset is not None
            array = datetime_to_numeric(array, offset, datetime_unit="us")

    if nax == 1 and by_.ndim > 1 and expected_ is None:
        # When we reduce along all axes, we are guaranteed to see all
        # groups in the final combine stage, so everything works.
        # This is not necessarily true when reducing along a subset of axes
        # (of by)
        # TODO: Does this depend on chunking of by?
        # For e.g., we could relax this if there is only one chunk along all
        # by dim != axis?
        raise NotImplementedError("Please provide ``expected_groups`` when not reducing along all axes.")

    assert nax <= by_.ndim
    if nax < by_.ndim:
        by_ = _move_reduce_dims_to_end(by_, tuple(-array.ndim + ax + by_.ndim for ax in axis_))
        array = _move_reduce_dims_to_end(array, axis_)
        axis_ = tuple(array.ndim + np.arange(-nax, 0))
        nax = len(axis_)

    # When axis is a subset of possible values; then npg will
    # apply the fill_value to groups that don't exist along a particular axis (for e.g.)
    # since these count as a group that is absent. thoo!
    # fill_value applies to all-NaN groups as well as labels in expected_groups that are not found.
    #     The only way to do this consistently is mask out using min_count
    #     Consider np.sum([np.nan]) = np.nan, np.nansum([np.nan]) = 0
    if min_count is None:
        if nax < by_.ndim or (fill_value is not None and provided_expected):
            min_count_: int = 1
        else:
            min_count_ = 0
    else:
        min_count_ = min_count

    # TODO: set in xarray?
    if min_count_ > 0 and func in ["nansum", "nanprod"] and fill_value is None:
        # nansum, nanprod have fill_value=0, 1
        # overwrite than when min_count is set
        fill_value = np.nan

    kwargs = dict(axis=axis_, fill_value=fill_value)
    agg = _initialize_aggregation(func, dtype, array.dtype, fill_value, min_count_, finalize_kwargs)

    # Need to set this early using `agg`
    # It cannot be done in the core loop of chunk_reduce
    # since we "prepare" the data for flox.
    kwargs["engine"] = _choose_engine(by_, agg) if engine is None else engine

    groups: tuple[np.ndarray | DaskArray, ...]
    if has_cubed:
        if method is None:
            method = "map-reduce"

        if method not in ("map-reduce", "blockwise"):
            raise NotImplementedError(
                "Reduction for Cubed arrays is only implemented for methods 'map-reduce' and 'blockwise'."
            )

        from .cubed import cubed_groupby_agg

        partial_agg = partial(cubed_groupby_agg, **kwargs)

        result, groups = partial_agg(
            array=array,
            by=by_,
            expected_groups=expected_,
            agg=agg,
            reindex=reindex,
            method=method,
            sort=sort,
        )

        return (result, groups)

    elif not has_dask:
        reindex.set_blockwise_for_numpy()
        results = _reduce_blockwise(
            array,
            by_,
            agg,
            expected_groups=expected_,
            reindex=reindex,
            sort=sort,
            **kwargs,
        )
        groups = (results["groups"],)
        result = results[agg.name]

    else:
        if TYPE_CHECKING:
            # TODO: How else to narrow that array.chunks is there?
            assert isinstance(array, DaskArray)

        if (not any_by_dask and method is None) or method == "cohorts":
            preferred_method, chunks_cohorts = find_group_cohorts(
                by_,
                [array.chunks[ax] for ax in range(-by_.ndim, 0)],
                expected_groups=expected_,
                # when provided with cohorts, we *always* 'merge'
                merge=(method == "cohorts"),
            )
        else:
            preferred_method = "map-reduce"
            chunks_cohorts = {}

        method = _choose_method(method, preferred_method, agg, by_, nax)

        if agg.chunk[0] is None and method != "blockwise":
            raise NotImplementedError(
                f"Aggregation {agg.name!r} is only implemented for dask arrays when method='blockwise'."
                f"Received method={method!r}"
            )

        if (
            _is_arg_reduction(agg)
            and method == "blockwise"
            and not all(nchunks == 1 for nchunks in array.numblocks[-nax:])
        ):
            raise NotImplementedError(
                "arg-reductions are not supported with method='blockwise', use 'cohorts' instead."
            )

        if nax != by_.ndim and method in ["blockwise", "cohorts"]:
            raise NotImplementedError(
                "Must reduce along all dimensions of `by` when method != 'map-reduce'."
                f"Received method={method!r}"
            )

        # TODO: clean this up
        reindex = _validate_reindex(
            reindex,
            func,
            method,
            expected_,
            any_by_dask,
            is_duck_dask_array(array),
            array.dtype,
        )

        if reindex.array_type is ReindexArrayType.SPARSE_COO:
            if not HAS_SPARSE:
                raise ImportError("Package 'sparse' must be installed to reindex to a sparse.COO array.")
            if not _is_reindex_sparse_supported_reduction(func):
                raise NotImplementedError(
                    f"Aggregation {func=!r} is not supported when reindexing to a sparse array. "
                    "Please raise an issue"
                )

        if TYPE_CHECKING:
            assert isinstance(reindex, ReindexStrategy)
            assert method is not None

        # TODO: just do this in dask_groupby_agg
        # we always need some fill_value (see above) so choose the default if needed
        if kwargs["fill_value"] is None:
            kwargs["fill_value"] = agg.fill_value[agg.name]

        from .dask import dask_groupby_agg

        partial_agg = partial(dask_groupby_agg, **kwargs)

        # if preferred method is already blockwise, no need to rechunk
        if preferred_method != "blockwise" and method == "blockwise" and by_.ndim == 1:
            _, array = rechunk_for_blockwise(array, axis=-1, labels=by_)

        result, groups = partial_agg(
            array=array,
            by=by_,
            expected_groups=expected_,
            agg=agg,
            reindex=reindex,
            method=method,
            chunks_cohorts=chunks_cohorts,
            sort=sort,
        )

        if sort and method != "map-reduce":
            assert len(groups) == 1
            sorted_idx = np.argsort(groups[0])
            # This optimization helps specifically with resampling
            if not _issorted(sorted_idx):
                result = result[..., sorted_idx]
                groups = (groups[0][sorted_idx],)

    if factorize_early:
        assert len(groups) == 1
        (groups_,) = groups
        # nan group labels are factorized to -1, and preserved
        # now we get rid of them by reindexing
        # First, for "blockwise", we can have -1 repeated in different blocks
        # This breaks the reindexing so remove those first.
        if method == "blockwise" and (mask := groups_ == -1).sum(axis=-1) > 1:
            result = result[..., ~mask]
            groups_ = groups_[..., ~mask]

        # This reindex also handles bins with no data
        result = reindex_(
            result,
            from_=groups_,
            to=expected_,
            fill_value=fill_value,
            array_type=ReindexArrayType.AUTO,  # just reindex the received array
        ).reshape(result.shape[:-1] + grp_shape)
        groups = final_groups

    if is_bool_array and (_is_minmax_reduction(func) or _is_first_last_reduction(func)):
        result = result.astype(bool)

    # Output of count has an int dtype.
    if requires_numeric and func != "count":
        if is_npdatetime:
            result = result.astype(datetime_dtype)
        elif is_cftime:
            asdelta = _to_pytimedelta(result, unit="us")
            nanmask = np.isnan(result)
            asdelta[nanmask] = datetime.timedelta(microseconds=0)
            result = asdelta + offset
            result[nanmask] = np.timedelta64("NaT")

    groups = map(
        lambda g: g.to_numpy() if isinstance(g, pd.Index) and not isinstance(g, pd.RangeIndex) else g, groups
    )
    return (result, *groups)



# Unlike the other aggregate_* submodules, this one simply defines a wrapper function
# because we run the groupby on the underlying dense data.

from collections.abc import Callable
from functools import partial
from typing import Any, TypeAlias

import numpy as np
import sparse

from flox.factorize import _factorize_multiple, factorize_
from flox.lib import _is_sparse_supported_reduction
from flox.xrdtypes import INF, NINF, _get_fill_value
from flox.xrutils import notnull


def nanadd(a, b):
    """
    Annoyingly, there is no numpy ufunc for nan-skipping elementwise addition
    unlike np.fmin, np.fmax :(

    From https://stackoverflow.com/a/50642947/1707127
    """
    ab = a + b
    return np.where(np.isnan(ab), np.where(np.isnan(a), b, a), ab)


CallableMap: TypeAlias = dict[str, Callable[[np.ndarray, np.ndarray], np.ndarray]]
BINARY_OPS: CallableMap = {
    "sum": np.add,
    "nansum": nanadd,
    "max": np.maximum,
    "nanmax": np.fmax,
    "min": np.minimum,
    "nanmin": np.fmin,
}
HYPER_OPS: CallableMap = {"sum": np.multiply, "nansum": np.multiply}
IDENTITY: dict[str, Any] = {
    "sum": 0,
    "nansum": 0,
    "prod": 1,
    "nanprod": 1,
    "max": NINF,
    "nanmax": NINF,
    "min": INF,
    "nanmin": INF,
}


def _sparse_agg(
    group_idx: np.ndarray,
    array: sparse.COO,
    func: str,
    engine: str,
    axis: int = -1,
    size: int | None = None,
    fill_value=None,
    dtype=None,
    **kwargs,
):
    """Wrapper function, that unwraps the underlying dense arrays, executes the groupby,
    and constructs the output sparse array."""
    from flox.aggregations import generic_aggregate

    if not isinstance(array, sparse.COO):
        raise ValueError("Sparse aggregations only supported for sparse.COO arrays")

    if not _is_sparse_supported_reduction(func):
        raise ValueError(f"{func} is unsupported for sparse arrays.")

    group_idx_subset = group_idx[array.coords[axis, :]]
    if array.ndim > 1:
        new_by = tuple(array.coords[:axis, :]) + (group_idx_subset,)
    else:
        new_by = (group_idx_subset,)
    codes, groups, shape = _factorize_multiple(
        new_by, expected_groups=(None,) * len(new_by), any_by_dask=False
    )
    # factorize again so we can construct a sparse result
    sparse_codes, sparse_groups, sparse_shape, _, sparse_size, _ = factorize_(codes, axes=(0,))

    dense_result = generic_aggregate(
        sparse_codes,
        array.data,
        func=func,
        engine=engine,
        dtype=dtype,
        size=sparse_size,
        fill_value=fill_value,
    )
    dense_counts = generic_aggregate(
        sparse_codes,
        array.data,
        # This counts is used to handle fill_value, so we need a count
        # of populated data, regardless of NaN value
        func="len",
        engine=engine,
        dtype=int,
        size=sparse_size,
        fill_value=0,
    )
    assert len(sparse_groups) == 1
    result_coords = np.stack(tuple(g[i] for g, i in zip(groups, np.unravel_index(*sparse_groups, shape))))

    full_shape = array.shape[:-1] + (size,)
    count = sparse.COO(coords=result_coords, data=dense_counts, shape=full_shape, fill_value=0)

    assert axis in (-1, array.ndim - 1)
    grouped_count = generic_aggregate(
        group_idx, group_idx, engine=engine, func="len", dtype=np.int64, size=size, fill_value=0
    )
    total_count = sparse.COO.from_numpy(
        np.expand_dims(grouped_count, tuple(range(array.ndim - 1))), fill_value=0
    )

    assert func in BINARY_OPS
    binop = BINARY_OPS[func]
    ident = _get_fill_value(array.dtype, IDENTITY[func])
    diff_count = total_count - count
    if (hyper_op := HYPER_OPS.get(func, None)) is not None:
        fill = hyper_op(diff_count, array.fill_value) if (diff_count > 0).any() else ident
    else:
        if "max" in func or "min" in func:
            # Note that fill_value for total_count, and count is 0.
            # So the fill_value for the `fill` result is the False branch i.e. `ident`
            fill = np.where(diff_count > 0, array.fill_value, ident)
        else:
            raise NotImplementedError

    result = sparse.COO(coords=result_coords, data=dense_result, shape=full_shape, fill_value=ident)
    with_fill = binop(result, fill)
    return with_fill


def nanlen(
    group_idx: np.ndarray,
    array: sparse.COO,
    engine: str,
    axis: int = -1,
    size: int | None = None,
    fill_value=None,
    dtype=None,
    **kwargs,
):
    new_array = sparse.COO(
        coords=array.coords,
        data=notnull(array.data),
        shape=array.shape,
        fill_value=notnull(array.fill_value),
    )
    return _sparse_agg(
        group_idx, new_array, func="sum", engine=engine, axis=axis, size=size, fill_value=0, dtype=dtype
    )


def mean(
    group_idx: np.ndarray,
    array: sparse.COO,
    engine: str,
    axis: int = -1,
    size: int | None = None,
    fill_value=None,
    dtype=None,
    **kwargs,
):
    sums = sum(
        group_idx, array, func="sum", engine=engine, axis=axis, size=size, fill_value=fill_value, dtype=dtype
    )
    counts = nanlen(
        group_idx, array, func="sum", engine=engine, axis=axis, size=size, fill_value=0, dtype=dtype
    )
    return sums / counts


def nanmean(
    group_idx: np.ndarray,
    array: sparse.COO,
    engine: str,
    axis: int = -1,
    size: int | None = None,
    fill_value=None,
    dtype=None,
    **kwargs,
):
    sums = sum(
        group_idx,
        array,
        func="nansum",
        engine=engine,
        axis=axis,
        size=size,
        fill_value=fill_value,
        dtype=dtype,
    )
    counts = nanlen(
        group_idx, array, func="sum", engine=engine, axis=axis, size=size, fill_value=0, dtype=dtype
    )
    return sums / counts


sum = partial(_sparse_agg, func="sum")
nansum = partial(_sparse_agg, func="nansum")
max = partial(_sparse_agg, func="max")
nanmax = partial(_sparse_agg, func="nanmax")
min = partial(_sparse_agg, func="min")
nanmin = partial(_sparse_agg, func="nanmin")



"""Reindexing functions for groupby operations.

This module provides functions for reindexing arrays during groupby operations.
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd

from . import xrdtypes
from .lib import dask_array_type, sparse_array_type
from .xrutils import isnull

if TYPE_CHECKING:
    from .core import T_Axis


class ReindexArrayType(Enum):
    """
    Enum describing which array type to reindex to.

    These are enumerated, rather than accepting a constructor,
    because we might want to optimize for specific array types,
    and because they don't necessarily have the same signature.

    For example, scipy.sparse.COO only supports a fill_value of 0.
    """

    AUTO = auto()
    NUMPY = auto()
    SPARSE_COO = auto()
    # Sadly, scipy.sparse.coo_array only supports fill_value = 0
    # SCIPY_SPARSE_COO = auto()
    # SPARSE_GCXS = auto()

    def is_same_type(self, other) -> bool:
        match self:
            case ReindexArrayType.AUTO:
                return True
            case ReindexArrayType.NUMPY:
                return isinstance(other, np.ndarray)
            case ReindexArrayType.SPARSE_COO:
                import sparse

                return isinstance(other, sparse.COO)


@dataclass
class ReindexStrategy:
    """
    Strategy for reindexing.

    Attributes
    ----------
    blockwise: bool, optional
        Whether to reindex at the blockwise step. Must be False for method="cohorts"
    array_type: ReindexArrayType, optional
        Whether to reindex to a different array type than array being reduced.
    """

    # whether to reindex at the blockwise step
    blockwise: bool | None
    array_type: ReindexArrayType = ReindexArrayType.AUTO

    def __post_init__(self):
        if self.blockwise is True:
            if self.array_type not in (ReindexArrayType.AUTO, ReindexArrayType.NUMPY):
                raise ValueError("Setting reindex.blockwise=True not allowed for non-numpy array type.")

    def set_blockwise_for_numpy(self):
        self.blockwise = True if self.blockwise is None else self.blockwise

    def get_dask_meta(self, other, *, fill_value, dtype) -> Any:
        if self.array_type is ReindexArrayType.AUTO:
            other = other._meta if isinstance(other, dask_array_type) else other
            if isinstance(other, sparse_array_type):
                return type(other).from_numpy(np.array([], dtype=dtype))
            return type(other)([], dtype=dtype)
        elif self.array_type is ReindexArrayType.NUMPY:
            return np.ndarray([], dtype=dtype)
        elif self.array_type is ReindexArrayType.SPARSE_COO:
            import sparse

            return sparse.COO.from_numpy(np.ones(shape=(0,) * other.ndim, dtype=dtype), fill_value=fill_value)


def reindex_numpy(array, from_: pd.Index, to: pd.Index, fill_value, dtype, axis: int):
    idx = from_.get_indexer(to)
    indexer: list[slice | np.ndarray] = [slice(None, None)] * array.ndim
    indexer[axis] = idx
    reindexed = array[tuple(indexer)]
    if (idx == -1).any():
        if fill_value is None:
            raise ValueError("Filling is required. fill_value cannot be None.")
        indexer[axis] = idx == -1
        reindexed = reindexed.astype(dtype, copy=False)
        reindexed[tuple(indexer)] = fill_value
    return reindexed


def reindex_pydata_sparse_coo(array, from_: pd.Index, to: pd.Index, fill_value, dtype, axis: int):
    import sparse

    assert axis == -1

    # Are there any elements in `to` that are not in `from_`.
    if isinstance(to, pd.RangeIndex) and len(to) > len(from_):
        # 1. pandas optimizes set difference between two RangeIndexes only
        # 2. We want to avoid realizing a very large numpy array in to memory.
        #    This happens in the `else` clause.
        #    There are potentially other tricks we can play, but this is a simple
        #    and effective one. If a user is reindexing to sparse, then len(to) is
        #    almost guaranteed to be > len(from_). If len(to) <= len(from_), then realizing
        #    another array of the same shape should be fine.
        needs_reindex = True
    else:
        needs_reindex = (from_.get_indexer(to) == -1).any()

    if needs_reindex and fill_value is None:
        raise ValueError("Filling is required. fill_value cannot be None.")

    idx = to.get_indexer(from_)
    mask = idx != -1  # indices along last axis to keep
    if mask.all():
        mask = slice(None)
    shape = array.shape

    if isinstance(array, sparse.COO):
        subset = array[..., mask]
        data = subset.data
        coords = subset.coords
        if subset.nnz > 0:
            coords[-1, :] = idx[mask][coords[-1, :]]
        if fill_value is None:
            # no reindexing is actually needed (dense case)
            # preserve the fill_value
            fill_value = array.fill_value
    else:
        ranges = np.broadcast_arrays(
            *np.ix_(*(tuple(np.arange(size) for size in shape[:axis]) + (idx[mask],)))
        )
        coords = np.stack(ranges, axis=0).reshape(array.ndim, -1)
        data = array[..., mask].reshape(-1)

    reindexed = sparse.COO(
        coords=coords,
        data=data.astype(dtype, copy=False),
        shape=(*array.shape[:axis], to.size),
        fill_value=fill_value,
    )

    return reindexed


def reindex_(
    array: np.ndarray,
    from_,
    to,
    *,
    array_type: ReindexArrayType = ReindexArrayType.AUTO,
    fill_value: Any = None,
    axis: T_Axis = -1,
    promote: bool = False,
) -> np.ndarray:
    if not isinstance(to, pd.Index):
        if promote:
            to = pd.Index(to)
        else:
            raise ValueError("reindex requires a pandas.Index or promote=True")

    if to.ndim > 1:
        raise ValueError(f"Cannot reindex to a multidimensional array: {to}")

    if array.shape[axis] == 0:
        # all groups were NaN
        shape = array.shape[:-1] + (len(to),)
        if array_type in (ReindexArrayType.AUTO, ReindexArrayType.NUMPY):
            reindexed = np.full(shape, fill_value, dtype=array.dtype)
        else:
            raise NotImplementedError
        return reindexed

    from_ = pd.Index(from_)
    # short-circuit for trivial case
    if from_.equals(to) and array_type.is_same_type(array):
        return array

    if from_.dtype.kind == "O" and isinstance(from_[0], tuple):
        raise NotImplementedError(
            "Currently does not support reindexing with object arrays of tuples. "
            "These occur when grouping by multi-indexed variables in xarray."
        )
    # Use '==' instead of 'is', as Dask serialization can break identity checks.
    if fill_value == xrdtypes.NA or isnull(fill_value):
        new_dtype, fill_value = xrdtypes.maybe_promote(array.dtype)
    else:
        new_dtype = array.dtype

    if array_type is ReindexArrayType.AUTO:
        if isinstance(array, sparse_array_type):
            array_type = ReindexArrayType.SPARSE_COO
        else:
            # TODO: generalize here
            # Right now, we effectively assume NEP-18 I think
            array_type = ReindexArrayType.NUMPY

    if array_type is ReindexArrayType.NUMPY:
        reindexed = reindex_numpy(array, from_, to, fill_value, new_dtype, axis)
    elif array_type is ReindexArrayType.SPARSE_COO:
        reindexed = reindex_pydata_sparse_coo(array, from_, to, fill_value, new_dtype, axis)
    return reindexed


__all__ = [
    "ReindexArrayType",
    "ReindexStrategy",
    "reindex_",
    "reindex_numpy",
    "reindex_pydata_sparse_coo",
]



"""Cubed-specific functions for groupby operations.

This module provides Cubed-specific implementations for groupby operations.
"""

from __future__ import annotations

from collections.abc import Sequence
from functools import partial
from typing import TYPE_CHECKING, Any

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from .aggregations import Aggregation
    from .core import T_Axes, T_Engine, T_Method
    from .types import CubedArray, T_By

from .core import (
    _finalize_results,
    _get_chunk_reduction,
    _is_arg_reduction,
    _reduce_blockwise,
)
from .reindex import ReindexStrategy
from .xrutils import is_chunked_array


def cubed_groupby_agg(
    array: CubedArray,
    by: T_By,
    agg: Aggregation,
    expected_groups: pd.Index | None,
    reindex: ReindexStrategy,
    axis: T_Axes = (),
    fill_value: Any = None,
    method: T_Method = "map-reduce",
    engine: T_Engine = "numpy",
    sort: bool = True,
    chunks_cohorts=None,
) -> tuple[CubedArray, tuple[pd.Index | np.ndarray | CubedArray]]:
    import cubed
    import cubed.core.groupby

    # I think _tree_reduce expects this
    assert isinstance(axis, Sequence)
    assert all(ax >= 0 for ax in axis)

    if method == "blockwise":
        assert by.ndim == 1
        assert expected_groups is not None

        def _reduction_func(a, by, axis, start_group, num_groups):
            # adjust group labels to start from 0 for each chunk
            by_for_chunk = by - start_group
            expected_groups_for_chunk = pd.RangeIndex(num_groups)

            axis = (axis,)  # convert integral axis to tuple

            blockwise_method = partial(
                _reduce_blockwise,
                agg=agg,
                axis=axis,
                expected_groups=expected_groups_for_chunk,
                fill_value=fill_value,
                engine=engine,
                sort=sort,
                reindex=reindex,
            )
            out = blockwise_method(a, by_for_chunk)
            return out[agg.name]

        num_groups = len(expected_groups)
        result = cubed.core.groupby.groupby_blockwise(
            array, by, axis=axis, func=_reduction_func, num_groups=num_groups
        )
        groups = (expected_groups,)
        return (result, groups)

    else:
        inds = tuple(range(array.ndim))

        by_input = by

        # Unifying chunks is necessary for argreductions.
        # We need to rechunk before zipping up with the index
        # let's always do it anyway
        if not is_chunked_array(by):
            # chunk numpy arrays like the input array
            chunks = tuple(array.chunks[ax] if by.shape[ax] != 1 else (1,) for ax in range(-by.ndim, 0))

            by = cubed.from_array(by, chunks=chunks, spec=array.spec)
        _, (array, by) = cubed.core.unify_chunks(array, inds, by, inds[-by.ndim :])

        # Cubed's groupby_reduction handles the generation of "intermediates", and the
        # "map-reduce" combination step, so we don't have to do that here.
        # Only the equivalent of "_simple_combine" is supported, there is no
        # support for "_grouped_combine".
        labels_are_unknown = is_chunked_array(by_input) and expected_groups is None
        do_simple_combine = not _is_arg_reduction(agg) and not labels_are_unknown

        assert do_simple_combine
        assert method == "map-reduce"
        assert expected_groups is not None
        assert reindex.blockwise is True
        assert len(axis) == 1  # one axis/grouping

        def _groupby_func(a, by, axis, intermediate_dtype, num_groups):
            blockwise_method = partial(
                _get_chunk_reduction(agg.reduction_type),
                func=agg.chunk,
                fill_value=agg.fill_value["intermediate"],
                dtype=agg.dtype["intermediate"],
                reindex=reindex,
                user_dtype=agg.dtype["user"],
                axis=axis,
                expected_groups=expected_groups,
                engine=engine,
                sort=sort,
            )
            out = blockwise_method(a, by)
            # Convert dict to one that cubed understands, dropping groups since they are
            # known, and the same for every block.
            return {f"f{idx}": intermediate for idx, intermediate in enumerate(out["intermediates"])}

        def _groupby_combine(a, axis, dummy_axis, dtype, keepdims):
            # this is similar to _simple_combine, except the dummy axis and concatenation is handled by cubed
            # only combine over the dummy axis, to preserve grouping along 'axis'
            dtype = dict(dtype)
            out = {}
            for idx, combine in enumerate(agg.simple_combine):
                field = f"f{idx}"
                out[field] = combine(a[field], axis=dummy_axis, keepdims=keepdims)
            return out

        def _groupby_aggregate(a, **kwargs):
            # Convert cubed dict to one that _finalize_results works with
            results = {"groups": expected_groups, "intermediates": a.values()}
            out = _finalize_results(results, agg, axis, expected_groups, reindex)
            return out[agg.name]

        # convert list of dtypes to a structured dtype for cubed
        intermediate_dtype = [(f"f{i}", dtype) for i, dtype in enumerate(agg.dtype["intermediate"])]
        dtype = agg.dtype["final"]
        num_groups = len(expected_groups)

        result = cubed.core.groupby.groupby_reduction(
            array,
            by,
            func=_groupby_func,
            combine_func=_groupby_combine,
            aggregate_func=_groupby_aggregate,
            axis=axis,
            intermediate_dtype=intermediate_dtype,
            dtype=dtype,
            num_groups=num_groups,
        )

        groups = (expected_groups,)

        return (result, groups)


__all__ = [
    "cubed_groupby_agg",
]
